from __future__ import annotations

from typing import Callable, Dict, List, Tuple

import cytoolz as tz
import numpy as np

from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.core.decorators import prerequisites
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation


@prerequisites(Spacy)
class LengthSubpopulation(ScoreSubpopulation):
    """Class to compute subpopulations based on text length."""

    def __init__(
        self,
        intervals: List[Tuple[int, int]],
        reduction_fn: Callable = np.sum,
        *args,
        **kwargs
    ):
        super(LengthSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=[
                Identifier(
                    _name=self.__class__.__name__,
                    gte=interval[0],
                    lte=interval[1],
                    reduction_fn=reduction_fn,
                )
                for interval in intervals
            ],
            *args,
            **kwargs,
        )

        # Assign the reduction fn
        self.reduction_fn = reduction_fn

    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        # Compute the length of each example under each key
        lengths = [
            Spacy.retrieve(
                batch=batch,
                columns=col,
                proc_fns=tz.compose(
                    # Compute lengths (# of words) for each tokenized text in a batch
                    lambda l: np.array([len(t) for t in l]),
                    # Extract tokens using Spacy
                    Spacy.tokens,
                ),
            )
            for col in columns
        ]

        # Reduction over the key axis
        return self.reduction_fn(np.array(lengths), axis=0)



CACHEDOPS = "cachedoperations"
SLICEBUILDERS = "slicebuilders"

GENERIC = "slicebuilder"
SUBPOPULATION = "subpopulation"
ATTACK = "attack"
AUGMENTATION = "augmentation"
TRANSFORMATION = "transformation"
CURATION = "curated"



import dill


class StorageMixin:
    """Mixin class for serialization."""

    def __init__(self, *args, **kwargs):
        super(StorageMixin, self).__init__(*args, **kwargs)

    def save(self, path: str) -> None:
        """Save the object."""
        dill.dump(self, open(path, "wb"))

    @classmethod
    def load(cls, path: str) -> object:
        """Load the object from the path."""
        return dill.load(open(path, "rb"))



import json
from typing import Callable, List, Optional, Tuple

import cytoolz as tz
import numpy as np

from robustnessgym.core.constants import TRANSFORMATION
from robustnessgym.core.dataset import Batch, Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.slicebuilder import SliceBuilder


class Transformation(SliceBuilder):
    def __init__(
        self,
        num_transformed: int = None,
        identifiers: List[Identifier] = None,
        apply_fn: Callable = None,
        category: str = None,
    ):
        assert (
            num_transformed if not identifiers else True
        ), "Must pass in num_transformed if no identifiers are given."

        super(Transformation, self).__init__(
            category=category if category else TRANSFORMATION,
            identifiers=[
                Identifier(
                    _name=f"{self.__class__.__name__}-{i + 1}",
                )
                for i in range(num_transformed)
            ]
            if not identifiers
            else identifiers,
            apply_fn=apply_fn,
        )

    @property
    def num_transformed(self):
        return self.num_slices

    def apply(
        self,
        skeleton_batches: List[Batch],
        slice_membership: np.ndarray,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Tuple[List[Batch], np.ndarray]:
        raise NotImplementedError

    def process_batch(
        self,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Tuple[List[Batch], Optional[np.ndarray]]:

        # Determine the size of the batch
        batch_size = len(batch[list(batch.keys())[0]])

        # Construct the matrix of slice labels: (batch_size x num_slices)
        slice_membership = np.ones((batch_size, self.num_slices), dtype=np.int32)

        # Uncache the batch to construct the skeleton for transformed batches
        skeleton_batches = [
            Dataset.uncached_batch(batch) for _ in range(self.num_slices)
        ]

        # Set the index for the skeleton batches
        for j, skeleton_batch in enumerate(skeleton_batches):
            skeleton_batch["index"] = [
                f"{idx}-{self.identifiers[j]}" for idx in skeleton_batch["index"]
            ]

        # Apply the SliceBuilder's core functionality
        transformed_batches, slice_membership = self.apply(
            skeleton_batches=skeleton_batches,
            slice_membership=slice_membership,
            batch=batch,
            columns=columns,
            *args,
            **kwargs,
        )

        # # Store the transformed examples
        # updates = self.construct_updates(
        #     transformed_batches=transformed_batches,
        #     slice_membership=slice_membership,
        #     batch_size=batch_size,
        #     columns=columns,
        #     mask=mask,
        #     compress=store_compressed,
        # )

        # Remove transformed examples where slice_membership[i, :] = 0 before returning
        transformed_batches = [
            self.filter_batch_by_slice_membership(
                batch=transformed_batch,
                slice_membership=slice_membership[:, j : j + 1],
            )[0]
            for j, transformed_batch in enumerate(transformed_batches)
        ]

        # if store:
        #     batch = self.store(
        #         batch=batch,
        #         updates=updates,
        #     )

        return transformed_batches, slice_membership

    def construct_updates(
        self,
        transformed_batches: List[Batch],
        slice_membership: np.ndarray,
        batch_size: int,
        columns: List[str],
        mask: List[int] = None,
        compress: bool = True,
    ):

        if compress:
            return [
                {
                    self.category: {
                        self.__class__.__name__: {
                            json.dumps(columns)
                            if len(columns) > 1
                            else columns[0]: [
                                tz.valmap(lambda v: v[i], transformed_batch)
                                for j, transformed_batch in enumerate(
                                    transformed_batches
                                )
                                if slice_membership[i, j]
                            ]
                        }
                    }
                }
                if np.any(slice_membership[i, :])
                else {}
                for i in range(batch_size)
            ]

        return [
            {
                self.category: {
                    self.__class__.__name__: {
                        str(self.identifiers[j]): {
                            json.dumps(columns)
                            if len(columns) > 1
                            else columns[0]: tz.valmap(
                                lambda v: v[i], transformed_batch
                            )
                        }
                        for j, transformed_batch in enumerate(transformed_batches)
                        if (not mask or not mask[j]) and (slice_membership[i, j])
                    }
                }
            }
            if np.any(slice_membership[i, :])
            else {}
            for i in range(batch_size)
        ]


class SingleColumnTransformation(Transformation):
    def single_column_apply(self, column_batch: List) -> List[List]:
        raise NotImplementedError

    def apply(
        self,
        skeleton_batches: List[Batch],
        slice_membership: np.ndarray,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Tuple[List[Batch], np.ndarray]:

        # Independently apply the transformation over the columns
        for column in columns:
            try:
                # Apply
                transformed_batch = self.single_column_apply(
                    column_batch=batch[column],
                )

                assert len(transformed_batch) == len(
                    batch[column]
                ), "Must output one list of augmentations per example."

                # Store the transformed text in the skeleton batches
                for i in range(slice_membership.shape[0]):
                    for j, transformed in enumerate(transformed_batch[i]):
                        skeleton_batches[j][column][i] = transformed
            except:  # noqa
                # Unable to transform: set all slice membership labels to zero
                slice_membership[:, :] = 0

        return skeleton_batches, slice_membership



from __future__ import annotations

import json
import logging
import os
import pathlib
from contextlib import contextmanager
from copy import copy, deepcopy
from typing import Callable, Dict, List, Mapping, Optional, Sequence, Union

import cytoolz as tz
import datasets
import pandas as pd
import torch
from datasets import Features
from jsonlines import jsonlines
from pyarrow import json as jsonarrow
from pyarrow import table

from robustnessgym.core.dataformats.inmemory import InMemoryDataset
from robustnessgym.core.dataformats.vision import VisionDataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.tape import InteractionTapeHierarchyMixin

logger = logging.getLogger(__name__)

Example = Dict
Batch = Dict[str, List]
BatchOrDataset = Union[Batch, "Dataset"]


class Dataset(
    InteractionTapeHierarchyMixin,
):
    """RobustnessGym Dataset class."""

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "robustnessgym/datasets/"

    # Create a directory
    logdir.mkdir(parents=True, exist_ok=True)

    def __init__(
        self,
        *args,
        identifier: Identifier = None,
        dataset_fmt: str = None,
        **kwargs,
    ):

        logger.debug("Creating Dataset.")

        # Internally, keep track of data inside a dataset
        self._dataset = None

        # Set a default dataset_fmt
        if len(args) == 1 and isinstance(args[0], datasets.Dataset):
            # If no `dataset_fmt` is passed in with a datasets.Dataset, just assume
            # `dataset_fmt` should be 'datasets'
            dataset_fmt = "datasets" if dataset_fmt is None else dataset_fmt
        else:
            # Default to `in_memory`, unless a dataset_fmt is passed in explicitly
            dataset_fmt = "in_memory" if dataset_fmt is None else dataset_fmt

        # InMemoryDataset
        if dataset_fmt == "in_memory":
            if len(args) == 1 and isinstance(args[0], InMemoryDataset):
                # Assign the dataset directly
                self._dataset = args[0]
            else:
                # Assign the dataset after converting to an InMemoryDataset
                self._dataset = InMemoryDataset(*args, **kwargs)

        # datasets.Dataset
        elif dataset_fmt == "datasets":
            if len(args) == 1 and isinstance(args[0], datasets.Dataset):
                # Assign the dataset directly
                self._dataset = args[0]
            else:
                # Assign the dataset after converting to a datasets.Dataset
                self._dataset = datasets.Dataset(*args, **kwargs)

        elif dataset_fmt == "image":
            self._dataset = VisionDataset(*args, **kwargs)

        else:
            raise NotImplementedError(
                "`dataset_fmt` must be one of ['in_memory', 'datasets', 'image']."
            )

        # Store the dataset format
        self._dataset_fmt = dataset_fmt

        # Call the InteractionTapeHierarchyMixin constructor
        InteractionTapeHierarchyMixin.__init__(self)

        # Create an identifier
        self._identifier = (
            self._autobuild_identifier() if not identifier else identifier
        )

        # Create logging directory
        self._create_logdir()

        # Add an index to the dataset
        if not self.has_index:
            self._add_index()

    @property
    def identifier(self):
        """Identifier."""
        return self._identifier

    @property
    def dataset_fmt(self):
        """Dataset format, one of ['in_memory', 'datasets', 'image']."""
        return self._dataset_fmt

    @property
    def features(self):
        """Dataset features."""
        return self._dataset.features

    @property
    def info(self):
        """Dataset info."""
        return self._dataset.info

    @property
    def split(self):
        """Dataset split."""
        return self._dataset.split

    @property
    def num_rows(self):
        """Number of rows in the dataset."""
        return self._dataset.num_rows

    @contextmanager
    def format(self, columns: List[str] = None):
        """Context where only `columns` will be visible."""
        # Get the current format
        current_format = self.get_format()

        if columns:
            # View only `columns`
            self.set_format(columns)
        else:
            # Use all columns
            self.set_format(self.column_names)
        try:
            yield
        finally:
            # Reset the format back
            self.set_format(current_format)

    def get_format(self) -> List[str]:
        """Get the dataset format."""
        return self._dataset.visible_columns

    def set_format(self, columns: List[str]):
        """Set the dataset format."""
        # TODO(karan): change `cache`
        if "cache" in self.column_names:
            return self._dataset.set_format(columns=columns + ["cache"])
        return self._dataset.set_format(columns=columns)

    def reset_format(self):
        """Set the dataset format."""
        return self._dataset.reset_format()

    def set_visible_rows(self, indices: Sequence):
        """Set the visible rows in the dataset."""
        self._dataset.set_visible_rows(indices)

    def reset_visible_rows(self):
        """Reset to make all rows visible."""
        self._dataset.reset_visible_rows()

    def add_column(self, column: str, values: List):
        """Add a column to the dataset."""
        self._dataset.add_column(column, values)

    def remove_column(self, column: str) -> None:
        """Remove a column from the dataset."""
        self._dataset.remove_column(column)

    def append(
        self,
        example_or_batch: Union[Example, Batch],
    ) -> None:
        """Append a batch of data to the dataset.

        `example_or_batch` must have the same columns as the dataset
        (regardless of what columns are visible).
        """
        self._dataset.append(example_or_batch)

    def _add_index(self):
        """Add an index to the dataset."""
        self.add_column("index", [str(i) for i in range(len(self))])

    def head(self, n: int, columns: List[str] = None):
        """View the first `n` examples of the dataset."""
        with self.format(columns):
            return pd.DataFrame(self[:n])

    def _create_logdir(self):
        """Create and assign a directory for logging this dataset's files."""
        if self.identifier.name == "RGDataset":
            # TODO(karan): handle temporarily constructed datasets differently
            self.logdir /= str(self.identifier)
            self.logdir.mkdir(parents=True, exist_ok=True)
        else:
            self.logdir /= str(self.identifier)
            self.logdir.mkdir(parents=True, exist_ok=True)

    def _autobuild_identifier(self) -> Identifier:
        """Automatically build an identifier for the dataset using available
        information."""
        # Look for a name, otherwise assign a default
        _name = (
            self._dataset.info.builder_name
            if self._dataset.info.builder_name
            else "RGDataset"
        )

        # Check for split, version information
        split = str(self._dataset.split) if self._dataset.split else None
        version = str(self._dataset.version) if self._dataset.version else None

        # Add all available information to kwargs dict
        kwargs = {}
        if split:
            kwargs["split"] = split
        if version:
            kwargs["version"] = version

        # Create identifier
        return Identifier(_name=_name, **kwargs)

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        return self._dataset[index]

    def __repr__(self):
        return (
            f"RG{self.__class__.__name__}["
            f"num_rows: {self.num_rows}]({self.identifier})"
        )

    @property
    def column_names(self):
        """Name of the columns in the dataset."""
        return self._dataset.column_names

    @property
    def has_index(self) -> bool:
        """Check if the dataset has an index column."""
        if self._dataset.column_names:
            return "index" in self._dataset.column_names
        # Just return True if the dataset is empty
        return True

    @classmethod
    def uncached_batch(cls, batch: Batch, copy=True) -> Batch:
        """Return batch with the "cache" and "slices" columns removed."""
        return tz.keyfilter(
            lambda k: k not in ["cache", "slices"], deepcopy(batch) if copy else batch
        )

    @classmethod
    def uncached_example(cls, example: Dict, copy=True) -> Dict:
        """Return example with the "cache" and "slices" columns removed."""
        return tz.keyfilter(
            lambda k: k not in ["cache", "slices"],
            deepcopy(example) if copy else example,
        )

    @classmethod
    def list_datasets(cls) -> List[str]:
        """List datasets on Huggingface datasets.

        Returns: list of datasets
        """
        return datasets.list_datasets()

    @classmethod
    def load_dataset(
        cls,
        *args,
        dataset_fmt: str = "in_memory",
        **kwargs,
    ):
        """Create a Dataset using Huggingface datasets.load_dataset(..). Loads
        any dataset available in Huggingface Dataset Hub.

        Use this instead of datasets.load_dataset, so

        dict_of_datasets = datasets.load_dataset('boolq')

        becomes

        dict_of_datasets = Dataset.load_dataset('boolq')
        """
        # Load the dataset
        dataset = datasets.load_dataset(*args, **kwargs)

        if isinstance(dataset, dict):
            return dict(
                map(
                    lambda t: (t[0], cls(t[1], dataset_fmt=dataset_fmt)),
                    dataset.items(),
                )
            )
        else:
            return cls(dataset, dataset_fmt=dataset_fmt)

    @classmethod
    def load_image_dataset(cls, *args, **kwargs):
        """Create a Dataset from a dictionary with paths to images and image
        metadata.

        Pass argument image_keys to indicate what are the keys of the
        columns with paths to images (default="image_file").
        """
        return cls(*args, dataset_fmt="image", **kwargs)

    @classmethod
    def from_datasets(
        cls,
        dataset: datasets.Dataset,
        identifier: Identifier = None,
        dataset_fmt: str = None,
    ) -> Dataset:
        """Create a Dataset from a Huggingface datasets.Dataset."""
        return cls(
            dataset,
            identifier=identifier,
            dataset_fmt=dataset_fmt,
        )

    @classmethod
    def from_jsonl(
        cls,
        json_path: str,
        identifier: Identifier = None,
        dataset_fmt: str = "in_memory",
    ) -> Dataset:
        """Load a dataset from a .jsonl file on disk, where each line of the
        json file consists of a single example."""

        if dataset_fmt == "in_memory":
            # Load the .jsonl file
            with open(json_path) as f:
                data = [json.loads(line) for line in f]

            return cls(
                data,
                identifier=identifier
                if identifier
                else Identifier("RGDataset", jsonl=json_path),
                dataset_fmt=dataset_fmt,
            )

        elif dataset_fmt == "datasets":
            # Use jsonarrow to directly load the json
            return cls(
                jsonarrow.read_json(json_path),
                identifier=identifier,
                dataset_fmt=dataset_fmt,
            )
        else:
            raise NotImplementedError

    @classmethod
    def from_batch(
        cls,
        batch: Batch,
        identifier: Identifier = None,
        dataset_fmt: str = "in_memory",
    ) -> Dataset:
        """Convert a batch to a Dataset."""

        if dataset_fmt == "in_memory":
            return cls(batch, identifier=identifier, dataset_fmt=dataset_fmt)
        elif dataset_fmt == "datasets":
            return cls(table(batch), identifier=identifier, dataset_fmt=dataset_fmt)
        else:
            raise NotImplementedError

    @classmethod
    def from_batches(
        cls,
        batches: Sequence[Batch],
        identifier: Identifier = None,
        dataset_fmt: str = "in_memory",
    ) -> Dataset:
        """Convert a list of batches to a dataset."""

        return cls.from_batch(
            tz.merge_with(
                tz.compose(list, tz.concat),
                *batches,
            ),
            identifier=identifier,
            dataset_fmt=dataset_fmt,
        )

    @classmethod
    def from_dict(
        cls,
        d: Dict,
        identifier: Identifier = None,
        dataset_fmt: str = "in_memory",
    ) -> Dataset:
        """Convert a dictionary to a dataset.

        Alias for Dataset.from_batch(..).
        """
        return cls.from_batch(
            batch=d,
            identifier=identifier,
            dataset_fmt=dataset_fmt,
        )

    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        identifier: Identifier = None,
        dataset_fmt: str = "in_memory",
    ):
        """Create a Dataset from a pandas DataFrame."""
        return cls.from_batch(
            df.to_dict("list"),
            identifier=identifier,
            dataset_fmt=dataset_fmt,
        )

    @classmethod
    def from_feather(
        cls,
        path: str,
        identifier: Identifier = None,
        dataset_fmt: str = "in_memory",
    ):
        """Create a Dataset from a feather file."""
        return cls.from_batch(
            pd.read_feather(path).to_dict("list"),
            identifier=Identifier("Feather", path=path)
            if not identifier
            else identifier,
            dataset_fmt=dataset_fmt,
        )

    def to_pandas(self) -> pd.DataFrame:
        """Convert a Dataset to a pandas DataFrame."""
        return pd.DataFrame(self[:])

    def to_dataloader(
        self,
        columns: Sequence[str],
        column_to_transform: Optional[Mapping[str, Callable]] = None,
        **kwargs,
    ) -> torch.utils.data.DataLoader:
        """Get a PyTorch dataloader that iterates over a subset of the columns
        (specified by `columns`) in the dataset. This is handy when using the dataset
        with training or evaluation loops outside of robustnessgym.  For example:
        ```
        dataset = Dataset(...)
        for img, target in dataset.to_dataloader(
            columns=["img_path", "label"],
            batch_size=16,
            num_workers=12
        ):
            out = model(target)
            loss = loss(out, target)
            ...
        ```

        Args:
            columns (Sequence[str]): A subset of the columns in the dataset.
                Specifies the columns to load. The dataloader will return values in same
                 order as `columns` here.
            column_to_transform (Optional[Mapping[str, Callable]], optional): A mapping
                from zero or more `columns` to callable transforms to be applied by the
                dataloader. Defaults to None, in which case no transforms are applied.
                e.g. `column_to_transform={"img_path": transforms.Resize((128,128))}`.

        Returns:
            torch.utils.data.DataLoader: dataloader that iterates over dataset
        """
        if not hasattr(self._dataset, "to_dataloader"):
            raise NotImplementedError(
                f'`to_dataloader` is not supported for format "{self._dataset_fmt}"'
            )
        return self._dataset.to_dataloader(
            columns=columns, column_to_transform=column_to_transform, **kwargs
        )

    def to_jsonl(self, path: str) -> None:
        """Save a Dataset to a jsonl file."""
        with jsonlines.open(path, mode="w") as writer:
            for example in self:
                writer.write(example)

    def batch(self, batch_size: int = 32):
        """Batch the dataset.

        Args:
            batch_size: integer batch size

        Returns:
        """
        for i in range(0, len(self), batch_size):
            yield self[i : i + batch_size]

    def update(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        # input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        # remove_columns: Optional[List[str]] = None,
        # keep_in_memory: bool = False,
        # load_from_cache_file: bool = True,
        # cache_file_name: Optional[str] = None,
        # writer_batch_size: Optional[int] = 1000,
        # features: Optional[Features] = None,
        # disable_nullable: bool = False,
        # fn_kwargs: Optional[dict] = None,
        # num_proc: Optional[int] = None,
        # suffix_template: str = "_{rank:05d}_of_{num_proc:05d}",
        # new_fingerprint: Optional[str] = None,
        **kwargs,
    ) -> Dataset:
        """Map on the dataset."""

        assert isinstance(self._dataset, InMemoryDataset) or isinstance(
            self._dataset, VisionDataset
        ), "Cannot apply .update() if format isn't InMemoryDataset."
        # Compute the map using the underlying dataset's .map()
        output = self._dataset.update(
            function=function,
            with_indices=with_indices,
            # input_columns=input_columns,
            batched=batched,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            # remove_columns=remove_columns,
            # keep_in_memory=keep_in_memory,
            # load_from_cache_file=load_from_cache_file,
            # cache_file_name=cache_file_name,
            # writer_batch_size=writer_batch_size,
            # features=features,
            # disable_nullable=disable_nullable,
            # fn_kwargs=fn_kwargs,
            # num_proc=num_proc,
            # suffix_template=suffix_template,
            # new_fingerprint=new_fingerprint,
            **kwargs,
        )

        if isinstance(output, datasets.Dataset):
            dataset = copy(self)
            dataset._dataset = output
        elif isinstance(output, InMemoryDataset):
            dataset = copy(self)
            dataset._dataset = output
        elif isinstance(output, VisionDataset):
            dataset = copy(self)
            dataset._dataset = output
        elif output is None:
            dataset = self
        else:
            raise NotImplementedError("Unrecognized dataset.")

        return dataset

    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        remove_columns: Optional[List[str]] = None,
        keep_in_memory: bool = False,
        load_from_cache_file: bool = True,
        cache_file_name: Optional[str] = None,
        writer_batch_size: Optional[int] = 1000,
        features: Optional[Features] = None,
        disable_nullable: bool = False,
        fn_kwargs: Optional[dict] = None,
        num_proc: Optional[int] = None,
        suffix_template: str = "_{rank:05d}_of_{num_proc:05d}",
        new_fingerprint: Optional[str] = None,
        **kwargs,
    ) -> Union[Dict, List]:
        """Map on the dataset."""

        assert isinstance(self._dataset, InMemoryDataset) or isinstance(
            self._dataset, VisionDataset
        ), (
            "Cannot apply .update() if format isn't InMemoryDataset or" "VisionDataset."
        )
        # Compute the map using the underlying dataset's .map()
        return self._dataset.map(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            batched=batched,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            remove_columns=remove_columns,
            keep_in_memory=keep_in_memory,
            load_from_cache_file=load_from_cache_file,
            cache_file_name=cache_file_name,
            writer_batch_size=writer_batch_size,
            features=features,
            disable_nullable=disable_nullable,
            fn_kwargs=fn_kwargs,
            num_proc=num_proc,
            suffix_template=suffix_template,
            new_fingerprint=new_fingerprint,
        )

    def filter(
        self,
        function: Optional[Callable] = None,
        with_indices=False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batch_size: Optional[int] = 1000,
        remove_columns: Optional[List[str]] = None,
        keep_in_memory: bool = False,
        load_from_cache_file: bool = True,
        cache_file_name: Optional[str] = None,
        writer_batch_size: Optional[int] = 1000,
        fn_kwargs: Optional[dict] = None,
        num_proc: Optional[int] = None,
        suffix_template: str = "_{rank:05d}_of_{num_proc:05d}",
        new_fingerprint: Optional[str] = None,
    ) -> Dataset:
        """Filter operation on the dataset."""
        # Compute the filter using the underlying dataset's .filter()
        output = self._dataset.filter(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            batch_size=batch_size,
            remove_columns=remove_columns,
            keep_in_memory=keep_in_memory,
            load_from_cache_file=load_from_cache_file,
            cache_file_name=cache_file_name,
            writer_batch_size=writer_batch_size,
            fn_kwargs=fn_kwargs,
            num_proc=num_proc,
            suffix_template=suffix_template,
            new_fingerprint=new_fingerprint,
        )

        if isinstance(output, datasets.Dataset):
            dataset = copy(self)
            dataset._dataset = output
        elif isinstance(output, InMemoryDataset):
            dataset = copy(self)
            dataset._dataset = output
        elif isinstance(output, VisionDataset):
            dataset = deepcopy(self)
            dataset._dataset = output
        else:
            raise NotImplementedError("Unrecognized dataset generated after .filter().")

        return dataset

    @classmethod
    def interleave(
        cls,
        datasets: List[Dataset],
        identifier: Identifier,
    ) -> Dataset:

        """Interleave a list of datasets."""
        return cls.from_batch(
            tz.merge_with(
                tz.compose(list, tz.interleave),
                *[dataset[:] for dataset in datasets],
            ),
            identifier=identifier,
        )

    @classmethod
    def chain(
        cls,
        datasets: List[Dataset],
        identifier: Identifier,
    ) -> Dataset:

        """Chain a list of datasets."""
        return cls.from_batch(
            tz.merge_with(
                tz.compose(list, tz.concat),
                *[dataset[:] for dataset in datasets],
            ),
            identifier=identifier,
        )

    @classmethod
    def load_from_disk(cls, path: str = None, identifier: Identifier = None) -> Dataset:
        """Load a dataset stored on disk."""
        assert (
            path or identifier and not (path and identifier)
        ), "Pass one of `path` or `identifier`."

        if identifier:
            # Use the default logdir to create a path to the dataset
            path = cls.logdir / str(identifier)
            if not os.path.exists(str(path)):
                raise OSError(f"Path {path} does not exist.")

        # Create an empty state
        state = {}

        # Load the metadata
        metadata = json.load(open(os.path.join(path, "metadata.json")))

        # Load the data
        if metadata["_dataset_fmt"] == "in_memory":
            state["_dataset"] = InMemoryDataset.load_from_disk(
                os.path.join(path, "_dataset")
            )
        elif metadata["_dataset_fmt"] == "datasets":
            state["_dataset"] = datasets.Dataset.load_from_disk(
                os.path.join(path, "_dataset")
            )
        else:
            raise NotImplementedError(
                f"`dataset_fmt` {metadata['_dataset_fmt']} not recognized."
            )

        # Merge the metadata with the state
        state = {**state, **metadata}

        # Create an empty dataset
        dataset = cls()
        dataset.__setstate__(state)

        return dataset

    def save_to_disk(self, path: str = None) -> None:
        """Save a dataset to disk."""
        if path is None:
            path = str(self.logdir)

        # Make all the directories to the path
        os.makedirs(path, exist_ok=True)

        # Get the dataset state
        state = self.__getstate__()

        # Save the data to disk
        os.makedirs(os.path.join(path, "_dataset"), exist_ok=True)
        state["_dataset"].save_to_disk(os.path.join(path, "_dataset"))

        # Save the metadata to disk
        json.dump(
            {k: v for k, v in state.items() if k != "_dataset"},
            open(os.path.join(path, "metadata.json"), "w"),
        )

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {
            "interactions",
            "_identifier",
            "_dataset",
            "_dataset_fmt",
        }

    @classmethod
    def _assert_state_keys(cls, state: Dict) -> None:
        """Assert that a state contains all required keys."""
        assert (
            set(state.keys()) == cls._state_keys()
        ), f"State must contain all state keys: {cls._state_keys()}."

    def __getstate__(self):
        """Get the current state of the dataset."""

        state = {
            # Interaction state
            "interactions": self.dumps_interactions(),
            # Identifier
            "_identifier": self.identifier.dumps() if self.identifier else None,
            # Dataset
            "_dataset": self._dataset,
            # Dataset format
            "_dataset_fmt": self._dataset_fmt,
        }
        Dataset._assert_state_keys(state)

        return state

    def __setstate__(self, state):
        """Set the current state of the dataset."""
        # Check that the state contains all keys
        Dataset._assert_state_keys(state)

        # Load the interactions
        self.interactions = self.loads_interactions(state["interactions"]).interactions

        # Load the identifier
        self._identifier = (
            Identifier.loads(state["_identifier"]) if state["_identifier"] else None
        )

        # Load the dataset
        self._dataset = state["_dataset"]

        # Set the dataset format
        self._dataset_fmt = state["_dataset_fmt"]

        # Update the logging directory
        self.logdir = Dataset.logdir / str(self.identifier)


def transpose_batch(batch: Batch):
    """Transpose a batch of data from a dict of lists to a list of dicts.

    Args:
        batch: batch of data which is a dictionary mapping columns to lists

    Returns: list of dicts, each dict corresponding to a single example
    """
    return [dict(zip(batch, t)) for t in zip(*batch.values())]



from typing import Dict, List, Optional

from torch import cuda

from robustnessgym.core.cachedops import SingleColumnCachedOperation
from robustnessgym.core.decorators import singlecolumn

try:
    from bootleg.end2end.bootleg_annotator import BootlegAnnotator
except ImportError:
    _bootleg_available = False
else:
    _bootleg_available = True


class Bootleg(SingleColumnCachedOperation):
    def __init__(
        self,
        config: Optional[Dict] = None,
        device: Optional[int] = None,
        cand_map: Optional[int] = None,
        threshold: Optional[float] = 0.5,
        cache_dir: Optional[str] = None,
        model_name: Optional[str] = None,
        *args,
        **kwargs
    ):
        if not _bootleg_available:
            raise ImportError(
                "Bootleg not available for import. Please see "
                "https://bootleg.readthedocs.io/en/latest/gettingstarted/install.html "
                "for help getting started."
            )
        # Pass threshold to SingleColumnCachedOperation to use as identifier
        super(Bootleg, self).__init__(threshold=threshold, *args, **kwargs)

        # Set the device
        if not device:
            device = "cuda" if cuda.is_available() else "cpu"

        # Create the annotator
        self.annotator = BootlegAnnotator(
            config=config,
            device=device,
            cand_map=cand_map,
            threshold=threshold,
            cache_dir=cache_dir,
            model_name=model_name,
        )

    @singlecolumn
    def apply(self, batch: Dict[str, List], columns: List[str], **kwargs) -> List:

        # Extract the test - Bootleg will automatically window sentences
        batch_text = batch[columns[0]]

        list_bootleg_results = [
            self.annotator.label_mentions(text) for text in batch_text
        ]

        return list_bootleg_results

    def encode(cls, obj) -> str:
        """Custom encode to allow for np.ndarray to be passed."""
        return obj

    def decode(cls, s: str):
        return s



"""Import common classes."""
# flake8: noqa
from robustnessgym.logging.utils import (
    initialize_logging,
    set_logging_level,
    set_logging_level_for_imports,
)

initialize_logging()

from robustnessgym.active.mandoline import mandoline
from robustnessgym.cachedops.allen.allen_predictor import AllenPredictor
from robustnessgym.cachedops.allen.constituency_parser import AllenConstituencyParser
from robustnessgym.cachedops.allen.dependency_parser import AllenDependencyParser
from robustnessgym.cachedops.allen.semantic_role_labeler import AllenSemanticRoleLabeler
from robustnessgym.cachedops.bootleg import Bootleg
from robustnessgym.cachedops.similarity import (
    RougeMatrix,
    RougeScore,
    SentenceSimilarityMatrix,
)
from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.cachedops.stanza import Stanza
from robustnessgym.cachedops.strip_text import StripText
from robustnessgym.cachedops.textblob import TextBlob
from robustnessgym.core.cachedops import (
    CachedOperation,
    SingleColumnCachedOperation,
    stow,
)
from robustnessgym.core.dataset import Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import Slice
from robustnessgym.core.testbench import DevBench, TestBench
from robustnessgym.slicebuilders.attacks.textattack import TextAttack
from robustnessgym.slicebuilders.slicebuilder import (
    SliceBuilder,
    SliceBuilderCollection,
)
from robustnessgym.slicebuilders.subpopulations.constituency_overlap import (
    ConstituencyOverlapSubpopulation,
    ConstituencySubtreeSubpopulation,
    FuzzyConstituencySubtreeSubpopulation,
)
from robustnessgym.slicebuilders.subpopulations.entity_frequency import EntityFrequency
from robustnessgym.slicebuilders.subpopulations.hans import (
    HansAdjectives,
    HansAdjectivesCompEnt,
    HansAdjectivesCompNonEnt,
    HansAdverbs,
    HansAdvsEntailed,
    HansAdvsNonEntailed,
    HansAllPhrases,
    HansCalledObjects,
    HansConjs,
    HansConstAdv,
    HansConstQuotEntailed,
    HansEntComplementNouns,
    HansFoodWords,
    HansIntransitiveVerbs,
    HansLocationNounsA,
    HansLocationNounsB,
    HansNonEntComplementNouns,
    HansNonEntQuotVerbs,
    HansNPSVerbs,
    HansNPZVerbs,
    HansPassiveVerbs,
    HansPastParticiples,
    HansPluralNouns,
    HansPluralNPZVerbs,
    HansPrepositions,
    HansQuestionEmbeddingVerbs,
    HansQuestions,
    HansReadWroteObjects,
    HansRelations,
    HansSingularNouns,
    HansToldObjects,
    HansTransitiveVerbs,
    HansUnderstoodArgumentVerbs,
    HansWonObjects,
)
from robustnessgym.slicebuilders.subpopulations.length import LengthSubpopulation
from robustnessgym.slicebuilders.subpopulations.lexical_overlap import (
    LexicalOverlapSubpopulation,
)
from robustnessgym.slicebuilders.subpopulations.phrase import (
    AhoCorasick,
    HasAllPhrases,
    HasAnyPhrase,
    HasComparison,
    HasDefiniteArticle,
    HasIndefiniteArticle,
    HasNegation,
    HasPhrase,
    HasPosessivePreposition,
    HasQuantifier,
    HasTemporalPreposition,
)
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation
from robustnessgym.slicebuilders.subpopulations.similarity import (
    Abstractiveness,
    Dispersion,
    Distillation,
    Ordering,
    Position,
    RougeMatrixScoreSubpopulation,
    RougeScoreSubpopulation,
)
from robustnessgym.slicebuilders.transformations.eda import EasyDataAugmentation
from robustnessgym.slicebuilders.transformations.fairseq import FairseqBacktranslation
from robustnessgym.slicebuilders.transformations.nlpaug import NlpAugTransformation
from robustnessgym.slicebuilders.transformations.similarity import (
    RougeMatrixSentenceTransformation,
)
from robustnessgym.tasks.task import (
    BinaryNaturalLanguageInference,
    BinarySentiment,
    ExtractiveQuestionAnswering,
    NaturalLanguageInference,
    QuestionAnswering,
    SentimentClassification,
    Summarization,
    Task,
    TernaryNaturalLanguageInference,
)

from .slicebuilders.attack import Attack
from .slicebuilders.curator import Curator
from .slicebuilders.subpopulation import Subpopulation, SubpopulationCollection



from collections import OrderedDict
from typing import Dict, List, Tuple

import cytoolz as tz
import numpy as np

try:
    import textattack.attack_recipes as attack_recipes
    from textattack.attack_recipes import AttackRecipe
    from textattack.models.wrappers import HuggingFaceModelWrapper, ModelWrapper
except ImportError:
    _textattack_available = False
    AttackRecipe, ModelWrapper = None, None
else:
    _textattack_available = True

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.model import Model
from robustnessgym.slicebuilders.attack import Attack


class TextAttack(Attack):
    """Class for TextAttack."""

    def __init__(
        self,
        attack: AttackRecipe,
    ):
        if not _textattack_available:
            raise ImportError("Textattack not found. Please `pip install textattack`.")

        super(TextAttack, self).__init__(
            identifiers=[
                Identifier(
                    self.__class__.__name__,
                    attack=attack,
                )
            ],
        )

        self.attack = attack

    @classmethod
    def recipes(cls):
        recipes = []
        for possible_recipe_name in dir(attack_recipes):
            possible_recipe = getattr(attack_recipes, possible_recipe_name)
            if hasattr(possible_recipe, "mro"):
                for _cls in possible_recipe.mro():
                    if _cls == AttackRecipe and possible_recipe != AttackRecipe:
                        recipes.append(possible_recipe_name)
        return recipes

    def apply(
        self,
        skeleton_batches: List[Dict[str, List]],
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> Tuple[List[Dict[str, List]], np.ndarray]:

        # Group the batch into inputs and output
        batch_inputs = tz.keyfilter(lambda k: k in columns[:-1], batch)
        batch_inputs = [
            OrderedDict(zip(batch_inputs, t)) for t in zip(*batch_inputs.values())
        ]

        batch_output = [int(e) for e in batch[columns[-1]]]

        # Create a fake dataset for textattack
        fake_dataset = list(zip(batch_inputs, batch_output))

        # Attack the dataset
        outputs = list(self.attack.attack_dataset(fake_dataset))

        for i, output in enumerate(outputs):
            # Check if the goal succeeded
            if output.perturbed_result.goal_status == 0:
                # If success, fill out the skeleton batch
                for (
                    key,
                    val,
                ) in output.perturbed_result.attacked_text._text_input.items():
                    # TODO(karan): support num_attacked_texts > 1
                    skeleton_batches[0][key][i] = val

                # # Fill the perturbed output: *this was incorrect, removing this
                # statement*
                # # TODO(karan): delete this snippet
                # skeleton_batches[0][columns[-1]][i] = output.perturbed_result.output
            else:
                # Unable to attack the example: set its slice membership to zero
                slice_membership[i, 0] = 0

        return skeleton_batches, slice_membership

    @classmethod
    def from_recipe(cls, recipe: str, model: ModelWrapper):
        return cls(attack=getattr(attack_recipes, recipe).build(model=model))

    @classmethod
    def wrap_huggingface_model(cls, model: Model) -> ModelWrapper:
        return HuggingFaceModelWrapper(model=model.model, tokenizer=model.tokenizer)



from types import SimpleNamespace

import dill as pickle
from semver import VersionInfo as Version


class SemanticVersionerMixin:
    """Simple mixin that adds semantic versioning to any class."""

    def __init__(self, version: str = "0.0.1", *args, **kwargs):
        super(SemanticVersionerMixin, self).__init__(*args, **kwargs)
        self._version = Version.parse(version)
        self._version_history = {}
        self._last_digest = None

        # TODO(karan): implement more features for commit-then-bump, add diffing

    @property
    def version(self):
        return str(self._version)

    @property
    def version_history(self):
        return self._version_history

    @property
    def major(self):
        return self._version.major

    @property
    def minor(self):
        return self._version.minor

    @property
    def patch(self):
        return self._version.patch

    def bump_major(self):
        """Commit the current version and bump the major version."""
        self.commit()
        self._version = self._version.bump_major()
        self._last_digest = self._digest()

    def bump_minor(self):
        """Commit the current version and bump the minor version."""
        self.commit()
        self._version = self._version.bump_minor()
        self._last_digest = self._digest()

    def bump_patch(self):
        """Commit the current version and bump the patch."""
        self.commit()
        self._version = self._version.bump_patch()
        self._last_digest = self._digest()

    def commit(self):
        """Commit the current version to history.

        Multiple commits on the same version overwrite each other.
        """
        self._version_history[self.version] = self._digest()

    def _digest(self) -> str:
        """Compute a digest for the object."""
        raise NotImplementedError(
            "Must implement a digest for the object that is being versioned."
        )

    def diff(self, digest: str, otherdigest: str) -> bool:
        """Check if digests have changed."""
        return digest != otherdigest

    def _dumps_version(self) -> str:
        return pickle.dumps(
            SimpleNamespace(
                version=self.version,
                history=self._version_history,
                last_digest=self._last_digest,
            )
        )

    def _loads_version(self, s: str):
        namespace = pickle.loads(s)
        self._version = namespace.version
        self._version_history = namespace.history
        self._last_digest = namespace.last_digest



from functools import partial
from typing import List, Sequence, Tuple

import numpy as np

from robustnessgym.cachedops.similarity import RougeMatrix
from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.core.dataset import Batch
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.tools import strings_as_json
from robustnessgym.slicebuilders.transformation import Transformation


class RougeMatrixSentenceTransformation(Transformation):
    def __init__(self, metric: Sequence[str], threshold: float):
        super(RougeMatrixSentenceTransformation, self).__init__(
            num_transformed=1,
            identifiers=Identifier.range(n=1, _name=self.__class__.__name__),
        )

        self.metric = metric
        self.threshold = threshold

    def apply(
        self,
        skeleton_batches: List[Batch],
        slice_membership: np.ndarray,
        batch: Batch,
        columns: List[str],
        *args,
        **kwargs
    ) -> Tuple[List[Batch], np.ndarray]:
        assert len(columns) == 2

        # Retrieve the relevant Rouge matrices
        matrices = RougeMatrix.retrieve(
            batch=batch,
            columns=columns,
            proc_fns=partial(RougeMatrix.select, metric=self.metric),
        )[strings_as_json(columns)]

        # Find max value along each row, remove rows that have max value below a
        # threshold
        rows_to_keep = [
            (m / (m.sum(axis=0) + 1e-5)).max(axis=1) >= self.threshold for m in matrices
        ]

        # Fetch sentences for the first column
        sentences = Spacy.retrieve(
            batch=batch,
            columns=[columns[0]],
            proc_fns=Spacy.sentences,
        )[columns[0]]

        # Delete sentences
        new_sentences = [
            " ".join(np.array(sent)[rows_to_keep[i]])
            for i, sent in enumerate(sentences)
        ]

        # Store the augmented text in the skeleton batches
        for i, augmented in enumerate(new_sentences):
            skeleton_batches[0][columns[0]][i] = augmented

        return skeleton_batches, slice_membership



from typing import Dict, List

import cytoolz as tz
import numpy as np

from robustnessgym.cachedops.spacy import Spacy
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation


class LexicalOverlapSubpopulation(ScoreSubpopulation):
    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        # Require that the number of keys is exactly 2
        assert len(columns) == 2, "Must specify exactly 2 keys."

        # Retrieve the tokens after lower-casing and placing into a set
        tokens = Spacy.retrieve(
            batch=batch,
            columns=[[key] for key in columns],
            proc_fns=tz.compose(
                # Lower case and put the tokens in a set for each tokenized text in
                # the batch
                lambda l: np.array(
                    [set([str(tok).lower() for tok in toks]) for toks in l]
                ),
                # Tokenize
                Spacy.tokens,
            ),
        )

        # Compute the intersection over union score
        return np.array(
            [
                len(tokens_0.intersection(tokens_1))
                / float(len(tokens_0.union(tokens_1)))
                for tokens_0, tokens_1 in zip(tokens[columns[0]], tokens[columns[1]])
            ]
        )



import math

import numpy as np
from sklearn.isotonic import IsotonicRegression


# Alternative to tree: isotonic regression
def fit_isotonic(y_test, prob_pos):
    """Given 0/1 labels 'y_test' and predicted model scores 'prob_pos', Train
    an isotonic regression model to calibrate model scores."""
    try:
        iso_reg = IsotonicRegression(out_of_bounds="clip").fit(prob_pos, y_test)
    except RuntimeError:
        iso_reg = None
    return iso_reg


def sample_iso(
    y_pred,
    prob_pos,
    iso_reg,
    samples,
    g,
    alpha,
    d_set,
):
    """
    Weighted sampling of rows to label
    Args:
        y_pred: predicted labels (k,)
        prob_pos: model scores (k,)
        iso_reg: isotonic regression model (or None)
        samples: number of datapoints to sample (int)
        g: current best-prediction of the F-Score (float in [0,1])
        alpha: defines the F_{\alpha} score estimated (float in [0,1])
            alpha = 0   -> recall
            alpha = 1   -> precision
            alpha = 0.5 -> F1
        d_set: indicators of whether each datapoint has already been labeled (k,)
            set this to np.zeros((k)) if no datapoints have been labeled
    Returns:
        sampled rows (samples,)
        weights of sample d rows (samples,)
    """
    num_rows = len(y_pred)

    # Deterministically sample the rows that have already been labeled
    # (since they require no additional budget)
    d_rows = np.where(d_set)[0]
    d_weights = np.ones(len(d_rows))
    if len(d_rows) > samples:
        return d_rows, d_weights
    samples -= len(d_rows)

    # Sample randomly if there's no isotonic regression model available for calibration
    # Otherwise, use the calibration model to compute calibrated probabilities 'p_1'
    if iso_reg is None:
        rand_rows = np.random.choice(num_rows, size=samples, replace=True)
        return np.concatenate([rand_rows, d_rows]), np.concatenate(
            [np.full((samples), 1.0 / num_rows), d_weights]
        )
    else:
        p_1 = iso_reg.predict(prob_pos) * 0.98 + 0.01  # Just to smooth 0 and 1

    # Compute Sawade et al.'s sampling weights from the calibrated probabilities 'p_1'
    weight_1 = np.sqrt(p_1 * ((1.0 - g) ** 2) + ((alpha * g) ** 2) * (1.0 - p_1))
    weight_0 = (1 - alpha) * np.sqrt(p_1 * (g ** 2))
    weights = weight_1 * y_pred + weight_0 * (1 - y_pred)

    # Sample according to 'weights' and return the rows sampled along with their
    # associated weights
    weight_sum = np.sum(weights)
    if weight_sum > 0:
        weights = weights[d_set is not True].squeeze()
        weights /= np.sum(weights)
        options = np.arange(num_rows)
        options = options[d_set is not True].squeeze()

        rand_row_indices = np.random.choice(
            len(options), size=samples, p=weights, replace=True
        )
        rand_rows = options[rand_row_indices]
        return np.concatenate([rand_rows, d_rows]), np.concatenate(
            [1.0 / (weights[rand_row_indices] * samples), d_weights]
        )
    else:
        # If all weights are zero, sample randomly
        rand_rows = np.random.choice(num_rows, size=samples, replace=True)
        return np.concatenate([rand_rows, d_rows]), np.concatenate(
            [np.full(samples, 1.0 / num_rows), d_weights]
        )


def get_fscore(y_pred, y_test, rows, weights):
    """
    Compute F-Score from sampled rows and weights
    Args:
        y_pred: model predictions (k,)
        y_test: ground-truth labels (k,)
        rows: sampled rows (samples,)
        weights: sampled weights (samples,)
    Returns:
        score: estimated F-Score
        trialstd: estimated standard deviation of F-Score estimate
        den:
    """
    unique_rows = np.unique(rows)
    unique_weights = np.array([weights[rows == elem].sum() for elem in unique_rows])
    rows = unique_rows
    weights = unique_weights

    # Compute F-Score
    num = ((y_test[rows] == y_pred[rows]) * weights).sum()
    den = weights.sum()
    score = num / den

    # Compute standard deviation of estimate
    var = ((((y_test[rows] == y_pred[rows]) - score) ** 2) * (weights ** 2)).mean() / (
        (weights.mean() ** 2) * (1.0 - ((weights ** 2).sum() / (weights.sum() ** 2)))
    )
    n = len(rows)
    trialstd = np.sqrt(var) / np.sqrt(n)

    return score, trialstd, den


def ais_singleiter(
    y_pred, y_test, prob_pos, sample_budget, g, alpha, known_rows, filter_rows
):
    """
    Perform a single AIS iteration of calibration + sampling
    Args:
        y_pred: model predictions (k,)
        y_test: ground-truth labels for sampled rows (samples,)
        prob_pos: model scores (k,)
        sample_budget: labeling budget for the iteration (int)
        g: current best-prediction of the F-Score (float in [0,1])
        alpha: defines the F_{\alpha} score estimated (float in [0,1])
            alpha = 0   -> recall
            alpha = 1   -> precision
            alpha = 0.5 -> F1
        known_rows: sampled rows (samples,)
        filtered_rows: indicator array allowing us to restrict sampling to a
        subset of rows (k,)
    Returns:
        score: estimated F-Score
        trialstd: estimated standard deviation of F-Score estimate
        den:
    """
    if np.sum(known_rows) > 0:
        iso_reg = fit_isotonic(y_test, prob_pos[known_rows])
    else:
        iso_reg = None
    rand_rows, weights = sample_iso(
        y_pred[filter_rows],
        prob_pos[filter_rows],
        iso_reg,
        sample_budget,
        g,
        alpha,
        known_rows[filter_rows],
    )
    return rand_rows, weights


def ais_fullalgorithm(y_pred, y_test, prob_pos, sample_budget):
    """
    Combine iterative AIS sampling with F-Score computation to compute best
    estimate of F-Score
    Args:
        y_pred: model predictions (k,)
        y_test: ground-truth labels (used as oracle) (k,)
        prob_pos: model scores (k,)
        sample_budget: total labeling budget (int)
    Returns:
        prf1: estimates of (precision, recall, F1)
        stds: estimates of standard deviation of estimated (precision, recall, F1)
        budget: actual labeling budget used (int)
    """

    # Initialize relevant variables
    k_keep = 4
    measurements = []
    stds = []
    measurement_weights = []
    all_rand_rows = []
    avg_budget = 0
    g = 0.5
    alpha = 0.5
    starting_budget = 10
    iteration_count = np.floor(np.log2(sample_budget / starting_budget)).astype(int) + 1

    # Sampling loop to iteratively sample batches of points to label
    for i in range(iteration_count):
        # Restrict sampling domain in early iterations when there aren't many
        # labeled positives
        # This significantly improves performance on rare categories
        poses = y_pred.sum()
        if (3 * (i + 1)) * poses < len(y_pred):
            filter_rows = np.argpartition(prob_pos, -((3 * (i + 1)) * poses))[
                -((3 * (i + 1)) * poses) :
            ]
        else:
            filter_rows = np.arange(len(y_pred))

        # Enumerate the already-sampled rows
        if len(all_rand_rows) > 0:
            unique_rows = np.unique(np.concatenate(all_rand_rows))
            known_rows = np.zeros(len(y_pred), dtype=bool)
            known_rows[unique_rows] = True
        else:
            known_rows = np.zeros(len(y_pred), dtype=bool)
        # Double sampling budget every iteration
        if i == (iteration_count - 1):
            iter_budget = sample_budget
        else:
            iter_budget = starting_budget * (2 ** i)

        # Use AIS algorithm to sample rows to label
        rand_rows, weights = ais_singleiter(
            y_pred=y_pred,
            y_test=y_test[known_rows],
            prob_pos=prob_pos,
            sample_budget=iter_budget,
            g=g,
            alpha=alpha,
            known_rows=known_rows,
            filter_rows=filter_rows,
        )
        all_rand_rows.append(filter_rows[rand_rows])
        rand_rows = filter_rows[rand_rows]
        weights *= len(rand_rows) / len(y_pred)

        # Compute precision, recall, and F1 using sampled rows and weights
        # Also computes the standard deviation of the estimates
        prec, prec_trialstd, prec_den = get_fscore(
            y_pred,
            y_test,
            rand_rows,
            weights * y_pred[rand_rows],
        )
        rec, rec_trialstd, rec_den = get_fscore(
            y_pred,
            y_test,
            rand_rows,
            weights * y_test[rand_rows],
        )
        f1, f1_trialstd, f1_den = get_fscore(
            y_pred,
            y_test,
            rand_rows,
            weights * (0.5 * y_pred[rand_rows] + 0.5 * y_test[rand_rows]),
        )
        measurements.append([prec, rec, f1])
        stds.append([prec_trialstd, rec_trialstd, f1_trialstd])

        # Update current best estimate of F1
        if not math.isnan(f1):
            g = 0.5 * g + 0.5 * f1
        measurement_weights.append([prec_den, rec_den, f1_den])

    all_rand_rows = np.unique(np.concatenate(all_rand_rows))
    measurements = np.asarray(measurements)
    stds = np.asarray(stds)
    measurement_weights = np.array(measurement_weights) + 0.0001

    # Keep only the results from the last 'k_keep' iterations of the algorithm
    if k_keep > 0:  # Set to -1 to deactivate
        measurements = measurements[-k_keep:]
        stds = stds[-k_keep:]
        measurement_weights = measurement_weights[-k_keep:]

    # Compute a weighted average of the estimates of F-Score computed across the
    # last 'k_keep' iterations
    avg_measurements = np.zeros(3)
    avg_stds = np.zeros(3)
    for k in range(3):
        indices = ~np.isnan(measurements[:, k])
        if indices.sum() > 0:
            normalized_weights = measurement_weights[:, k][indices]
            normalized_weights /= np.sum(normalized_weights)
            avg_measurements[k] = np.average(
                measurements[:, k][indices], weights=measurement_weights[:, k][indices]
            )
            avg_stds[k] = np.sqrt(
                np.sum((stds[:, k][indices] * normalized_weights) ** 2)
            )
        else:
            avg_measurements[k] = math.nan
            avg_stds[k] = math.nan
    avg_budget = len(all_rand_rows)

    return np.array(avg_measurements), np.array(avg_stds), np.array(avg_budget)



from robustnessgym.cachedops.allen.allen_predictor import AllenPredictor


class AllenConstituencyParser(AllenPredictor):
    def __init__(self, device: str = None, *args, **kwargs):
        super(AllenConstituencyParser, self).__init__(
            path="https://storage.googleapis.com/allennlp-public-models/elmo"
            "-constituency-parser-2020.02.10.tar.gz",
            device=device,
            *args,
            **kwargs,
        )

    @classmethod
    def encode(cls, prediction) -> str:
        # Extract the tree from the output of the constituency parser
        return super().encode(obj=prediction["trees"])



from robustnessgym.tasks.task import QuestionAnswering, TernaryNaturalLanguageInference

dataset_to_task = {
    # Natural Language Inference
    # Ternary
    "snli": TernaryNaturalLanguageInference,
    # Question Answering
    "squad": QuestionAnswering,
}



from typing import Dict, List

import fuzzywuzzy.fuzz as fuzz
import numpy as np

try:
    from nltk import Tree
except ImportError:
    _nltk_available = False
    Tree = None
else:
    _nltk_available = True

from robustnessgym.cachedops.allen.constituency_parser import AllenConstituencyParser
from robustnessgym.core.decorators import prerequisites
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation


@prerequisites(AllenConstituencyParser)
class ConstituencyOverlapSubpopulation(ScoreSubpopulation):
    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        # Require that the number of keys is exactly 2
        assert len(columns) == 2, "Must specify exactly 2 keys."

        # Retrieve the trees
        trees = AllenConstituencyParser.retrieve(
            batch=batch, columns=[[key] for key in columns]
        )
        trees_0, trees_1 = trees[columns[0]], trees[columns[1]]

        # Fuzzy match the trees and return the `scores`
        return np.array(
            [
                fuzz.partial_token_set_ratio(
                    tree_0.replace("(", "").replace(")", "").replace(" ", ""),
                    tree_1.replace("(", "").replace(")", "").replace(" ", ""),
                )
                for tree_0, tree_1 in zip(trees_0, trees_1)
            ]
        )


@prerequisites(AllenConstituencyParser)
class ConstituencySubtreeSubpopulation(ScoreSubpopulation):
    def __init__(self, *args, **kwargs):
        super(ConstituencySubtreeSubpopulation, self).__init__(
            intervals=[(1, 1)],
            identifiers=[Identifier(_name=self.__class__.__name__)],
            *args,
            **kwargs
        )

    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        # Require that the number of keys is exactly 2
        assert len(columns) == 2, "Must specify exactly 2 keys."

        # Retrieve the trees
        trees = AllenConstituencyParser.retrieve(
            batch=batch, columns=[[column] for column in columns]
        )
        trees_0, trees_1 = trees[columns[0]], trees[columns[1]]

        # Convert the trees corresponding to key 0 to NLTK trees
        trees_0 = [Tree.fromstring(tree) for tree in trees_0]

        # Find all subtrees of these trees
        all_subtrees_0 = [
            set(
                [
                    str(t).replace("\n", "").replace(" ", "").lower()
                    for t in tree_0.subtrees()
                ]
            )
            for tree_0 in trees_0
        ]

        # Output a score of 1 if the tree corresponding to key 1 lies in any subtree
        return np.array(
            [
                int(
                    tree_1.replace(" ", "")
                    .replace("(..)", "")
                    .replace("(,,)", "")
                    .lower()
                    in subtrees_0
                )
                for tree_1, subtrees_0 in zip(trees_1, all_subtrees_0)
            ]
        )


@prerequisites(AllenConstituencyParser)
class FuzzyConstituencySubtreeSubpopulation(ScoreSubpopulation):
    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        # Require that the number of keys is exactly 2
        assert len(columns) == 2, "Must specify exactly 2 keys."

        # Retrieve the trees
        trees = AllenConstituencyParser.retrieve(
            batch=batch, columns=[[column] for column in columns]
        )
        trees_0, trees_1 = trees[columns[0]], trees[columns[1]]

        # Convert the trees corresponding to key 0 to NLTK trees
        trees_0 = [Tree.fromstring(tree) for tree in trees_0]

        # Find all subtrees of these trees
        all_subtrees_0 = [
            set(
                [
                    str(t).replace("\n", "").replace(" ", "").lower()
                    for t in tree_0.subtrees()
                ]
            )
            for tree_0 in trees_0
        ]

        # Output a fuzzy score if the tree corresponding to key 1 is similar to any
        # subtree
        return np.array(
            [
                max(
                    [
                        fuzz.partial_ratio(
                            tree_1.replace(" ", "")
                            .replace("(..)", "")
                            .replace("(,,)", "")
                            .lower(),
                            subtree,
                        )
                        for subtree in subtrees_0
                    ]
                )
                for tree_1, subtrees_0 in zip(trees_1, all_subtrees_0)
            ]
        )



import urllib.request
import zipfile

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.tools import DownloadProgressBar
from robustnessgym.slicebuilders.subpopulations.phrase import HasAnyPhrase


class HasCategoryPhrase(HasAnyPhrase):
    def __init__(self):

        # Fetch wordlists
        self._fetch_sources()
        self.categories_to_words = self._load_all()

        super(HasCategoryPhrase, self).__init__(
            phrase_groups=[
                self.categories_to_words[supercategory][category]
                for (supercategory, category) in self.categories
            ],
            identifiers=[
                Identifier(
                    _name=self.__class__.__name__,
                    supercategory=supercategory,
                    category=category,
                )
                for (supercategory, category) in self.categories
            ],
        )

    @property
    def supercategories(self):
        return list(self.categories_to_words.keys())

    @property
    def categories(self):
        return sorted(
            [
                (supercategory, category)
                for supercategory in self.categories_to_words.keys()
                for category in self.categories_to_words[supercategory]
            ]
        )

    @classmethod
    def _fetch_sources(cls):
        if not (cls.logdir / "wordlists-master").exists():
            print("wordlists not found. Downloading..")
            urllib.request.urlretrieve(
                "https://github.com/imsky/wordlists/archive/master.zip",
                filename=str(cls.logdir / "wordlists.zip"),
                reporthook=DownloadProgressBar(),
            )

            with zipfile.ZipFile(str(cls.logdir / "wordlists.zip")) as zip_ref:
                zip_ref.extractall(str(cls.logdir))

    @classmethod
    def _load_all(cls):
        """Loads wordlists.

        Returns:
        """
        category_to_words = {
            supercategory: {} for supercategory in ["nouns", "verbs", "adjectives"]
        }
        for supercategory in ["nouns", "verbs", "adjectives"]:
            for path in (cls.logdir / "wordlists-master" / supercategory).glob("*"):
                with open(str(path)) as f:
                    category_to_words[supercategory][path.stem] = set(
                        f.read().splitlines()
                    )

        return category_to_words



from typing import Dict, List, Optional, Union

from robustnessgym.core.cachedops import SingleColumnCachedOperation
from robustnessgym.core.identifier import Identifier

try:
    import textblob
except ImportError:
    _textblob_available = False
else:
    _textblob_available = True


class TextBlob(SingleColumnCachedOperation):
    def __init__(self):
        if not _textblob_available:
            raise ImportError(
                "TextBlob not available for import. Install using "
                "\npip install textblob\npython -m textblob.download_corpora"
            )
        # TODO(karan): requires running `python -m textblob.download_corpora`
        super(TextBlob, self).__init__()

    @classmethod
    def encode(cls, obj: "textblob.TextBlob") -> str:
        # Dump the TextBlob object to JSON
        # This loses a lot of information
        # Unfortunately, TextBlob provides no way to serialize/deserialize objects
        return obj.to_json()

    @classmethod
    def retrieve(
        cls,
        batch: Dict[str, List],
        columns: List[str],
        identifier: Union[str, Identifier] = None,
        reapply: bool = False,
        **kwargs
    ) -> Optional[Dict[str, List]]:
        # Default to reapplying the TextBlob op when retrieving
        return super().retrieve(batch, columns, identifier, reapply=True, **kwargs)

    def single_column_apply(self, column_batch: List, *args, **kwargs) -> List:
        # Create a TextBlob for each example
        return [textblob.TextBlob(text) for text in column_batch]



from __future__ import annotations

import itertools
from functools import partial
from typing import Dict, List

import dill
import numpy as np
import pandas as pd
import plotly.figure_factory as ff
import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.subplots import make_subplots


class ReportColumn:
    """A single column in the Robustness Report."""

    def __init__(self, title: str):
        self.title = title

    def set_title(self, title: str):
        self.title = title


class ScoreColumn(ReportColumn):
    """A column for numeric scores in the Robustness Report, displayed as a bar
    chart."""

    def __init__(
        self, title: str, min_val: float, max_val: float, is_0_to_1: bool = False
    ):
        super(ScoreColumn, self).__init__(title)
        self.min_val = min_val
        self.max_val = max_val
        self.is_0_to_1 = is_0_to_1

    def set_min(self, min_val: float):
        self.min_val = min_val

    def set_max(self, max_val: float):
        self.max_val = max_val


class ClassDistributionColumn(ReportColumn):
    """A column for discrete class distributions in the Robustness Report,
    displayed as a heatmap."""

    def __init__(self, title: str, class_codes: List[str]):
        super(ClassDistributionColumn, self).__init__(title)
        self.class_codes = class_codes

    def set_class_codes(self, class_codes: List[str]):
        self.class_codes = class_codes


class NumericColumn(ReportColumn):
    """A column for numeric data in the Robustness Report, displayed as the raw
    value."""

    def __init__(self, title: str):
        super(NumericColumn, self).__init__(title)


class Report:
    """Class for Robustness Gym Report.
    Args:
        data: Pandas dataframe in the following format:
            column 1: category name
            column 2: slice name
            columns 3-N: data corresponding to passed columns parameter
        columns: ReportColumn objects specifying format of columns 3-N in data
        model_name (optional): model name to show in report
        dataset_name (optional): dataset name to show in report
        **kwargs (optional): any additional config paramters
    """

    def __init__(
        self,
        data: pd.DataFrame,
        columns: List[ReportColumn],
        model_name: str = None,
        dataset_name: str = None,
        **kwargs,
    ):

        # Make a copy of data since may be modified by methods below
        self.data = data.copy()

        self.columns = columns
        self.model_name = model_name
        self.dataset_name = dataset_name

        self.config = {
            "color_scheme": ["#ec7734", "#3499ec", "#ec34c1", "#9cec34"],
            "score_color_complement": "#F3F4F7",
            "text_fill_color": "#F3F4F7",
            "text_border_color": "#BEC4CE",
            "distribution_color_scale": [[0.0, "#FBF5F2"], [1.0, "#EC7734"]],
            "col_spacing": 0.035,
            "row_height": 24,
            "category_padding": 24,
            "header_padding": 80,
            "score_col_width": 0.6,
            "class_dist_col_width": 0.35,
            "numeric_col_width": 0.25,
            "layout_width": 960,
            "font_size_dist": 12,
            "font_size_data": 13,
            "font_size_heading": 14,
            "font_size_category": 14,
        }

        self.update_config(**kwargs)

    def sort(
        self, category_order: Dict[str, int] = None, slice_order: Dict[str, int] = None
    ):
        """Sort rows in report by category / slice alphabetically, or using
        specified order.

        Args:
          category_order (optional): map from category name to sorting rank. If None,
          sort categories alphabetically.
          slice_order (optional): map from slice name to sorting rank. If None, sort
          slices alphabetically (within a category).
        """

        if category_order is None:
            category_order = {}

        if slice_order is None:
            slice_order = {}

        for col_name in ["sort-order-category", "sort-order-slice"]:
            if col_name in self.data:
                raise ValueError(f"Column name '{col_name}' is reserved")

        self.data["sort-order-category"] = self.data[0].map(
            lambda x: (category_order.get(x, 2 ** 10000), x)
        )
        self.data["sort-order-slice"] = self.data[1].map(
            lambda x: (slice_order.get(x, 2 ** 10000), x)
        )

        self.data = self.data.sort_values(
            by=["sort-order-category", "sort-order-slice"]
        ).drop(["sort-order-category", "sort-order-slice"], axis="columns")

        self.data.reset_index(inplace=True, drop=True)

    def filter(self, categories: List[str] = None, slices: List[str] = None):
        """Filter report to specific categories AND slices
        Args:
          categories (optional): list of category names to filter by
          slices (optional):list of slice names to filter by
        """
        if categories is not None:
            # self.data = self.data.loc(self.data[0].isin(categories))
            self.data = self.data[self.data[0].isin(categories)]
        if slices is not None:
            self.data = self.data[self.data[1].isin(slices)]
        self.data.reset_index(inplace=True, drop=True)

    def rename(self, category_map: Dict[str, str], slice_map: Dict[str, str]):
        """Rename categories, slices
        Args:
            category_map (optional): map from old to new category name
            slice_map (optional): map from old to new slice name
        """
        if category_map is not None:
            self.data[0] = self.data[0].map(lambda x: category_map.get(x, x))
        if slice_map is not None:
            self.data[1] = self.data[1].map(lambda x: slice_map.get(x, x))

    def set_class_codes(self, class_cds: List[str]):
        """Set single-letter class codes used for class distribution
        columns."""
        for col in self.columns:
            if isinstance(col, ClassDistributionColumn):
                col.set_class_codes(class_cds)

    def set_model_name(self, model_name):
        """Set model name displayed on report."""
        self.model_name = model_name

    def set_dataset_name(self, dataset_name):
        """Set dataset name displayed on report."""
        self.dataset_name = dataset_name

    def set_range(self, col_title: str, min_val: float = None, max_val: float = None):
        """Set min and max values for score columns
        Args:
            col_title: title of column to update
            min_val: minimum value
            max_val: maximum value
        """
        for col in self.columns:
            if isinstance(col, ScoreColumn) and col.title == col_title:
                if min_val is not None:
                    col.min_val = min_val
                if max_val is not None:
                    col.max_val = max_val

    def update_config(self, **kwargs):
        for k, v in kwargs.items():
            if k not in self.config:
                raise ValueError(f"Invalid config param: '{k}'")
            self.config[k] = v

    def round(self):
        # Round everything
        self.data = self.data.round(3)
        self.data.class_dist = self.data.class_dist.apply(partial(np.round, decimals=3))
        self.data.pred_dist = self.data.pred_dist.apply(partial(np.round, decimals=3))

    @classmethod
    def load(cls, path: str) -> Report:
        obj = dill.load(open(path, "rb"))
        assert isinstance(obj, Report), (
            f"dill loaded an instance of {type(obj)}, " f"must load {cls.__name__}."
        )
        return obj

    def save(self, path: str):
        return dill.dump(self, open(path, "wb"))

    def figure(self, show_title=False) -> Figure:

        # Verify that rows are grouped by category
        row_categories = self.data[0].tolist()
        save_cat_groups = set()  # Previous category groupings already encountered
        prev_cat = None
        # Loop through each row and see if a category is encountered outside of first
        # identified group for that category
        for cat in row_categories:
            if cat != prev_cat:  # category changes
                if cat in save_cat_groups:  # if new category previously encountered
                    raise ValueError("Rows must be grouped by category.")
                prev_cat = cat
                save_cat_groups.add(cat)

        categories = []
        category_sizes = []  # Num rows in each category
        for category, group in itertools.groupby(self.data[0]):  # column 0 is category
            categories.append(category)
            category_sizes.append(len(list(group)))
        n_rows = sum(category_sizes)
        height = (
            n_rows * self.config["row_height"]
            + len(categories) * self.config["category_padding"]
            + self.config["header_padding"]
        )
        col_widths = []
        for col in self.columns:
            if isinstance(col, ScoreColumn):
                col_width = self.config["score_col_width"]
            elif isinstance(col, ClassDistributionColumn):
                col_width = self.config["class_dist_col_width"]
            else:
                col_width = self.config["numeric_col_width"]
            col_widths.append(col_width)

        fig = make_subplots(
            rows=len(categories),
            row_titles=categories,
            cols=len(self.columns),
            shared_yaxes=True,
            subplot_titles=[col.title for col in self.columns],
            horizontal_spacing=self.config["col_spacing"],
            vertical_spacing=self.config["category_padding"] / height,
            row_width=list(reversed(category_sizes)),
            column_width=col_widths,
        )

        hms = []
        coords = []
        category_ndx = 1
        # Group data by category
        for category, category_data in self.data.groupby(0, sort=False):
            score_col_ndx = 0
            slice_names = category_data[1]
            slice_names = [s + " " * 3 for s in slice_names]
            for col_ndx, col in enumerate(self.columns):
                df_col_ndx = col_ndx + 2
                # Dataframe has two leading columns with category, slice
                fig_col_ndx = col_ndx + 1  # figure columns are 1-indexed
                x = category_data[df_col_ndx].tolist()
                if isinstance(col, ScoreColumn):
                    if col.is_0_to_1:
                        x = [100 * x_i for x_i in x]
                    col_max = col.max_val
                    if col.is_0_to_1:
                        col_max = 100 * col.max_val
                    fig.add_trace(
                        go.Bar(
                            x=x,
                            y=slice_names,
                            orientation="h",
                            marker=dict(color=self.get_color(score_col_ndx)),
                            showlegend=False,
                            text=[f"{x_i:.1f}" for x_i in x],
                            textposition="inside",
                            width=0.95,
                            textfont=dict(color="white"),
                        ),
                        row=category_ndx,
                        col=fig_col_ndx,
                    )
                    # Add marker for gray fill
                    fig.add_trace(
                        go.Bar(
                            x=[col_max - x_i for x_i in x],
                            y=slice_names,
                            orientation="h",
                            marker=dict(color=self.config["score_color_complement"]),
                            showlegend=False,
                            width=0.9,
                        ),
                        row=category_ndx,
                        col=fig_col_ndx,
                    )
                    score_col_ndx += 1
                elif isinstance(col, ClassDistributionColumn):
                    annotation_text = [
                        [f"{int(round(z * 100)):d}" for z in rw] for rw in x
                    ]
                    hm = ff.create_annotated_heatmap(
                        x,
                        x=col.class_codes,
                        xgap=1,
                        ygap=1,
                        annotation_text=annotation_text,
                        colorscale=self.config["distribution_color_scale"],
                        zmin=0,
                        zmax=1,
                    )
                    hms.append(hm)
                    # Save annotation data for special code related to heatmaps at end
                    coords.append(len(self.columns) * (category_ndx - 1) + fig_col_ndx)
                    fig.add_trace(
                        hm.data[0],
                        row=category_ndx,
                        col=fig_col_ndx,
                    )
                elif isinstance(col, NumericColumn):
                    # Repurpose bar chart as text field.
                    fig.add_trace(
                        go.Bar(
                            x=[1] * len(x),
                            y=slice_names,
                            orientation="h",
                            marker=dict(
                                color=self.config["text_fill_color"],
                                line=dict(
                                    width=0, color=self.config["text_border_color"]
                                ),
                            ),
                            showlegend=False,
                            text=[human_format(x_i) for x_i in x],
                            textposition="inside",
                            insidetextanchor="middle",
                            width=0.9,
                        ),
                        row=category_ndx,
                        col=fig_col_ndx,
                    )
                else:
                    raise ValueError("Invalid col type")
            category_ndx += 1

        for category_ndx in range(1, len(categories) + 1):
            if category_ndx == len(categories):
                show_x_axis = True
            else:
                show_x_axis = False
            for col_ndx, col in enumerate(self.columns):
                fig_col_ndx = col_ndx + 1  # plotly cols are 1-indexed
                fig.update_yaxes(autorange="reversed", automargin=True)
                if isinstance(col, ScoreColumn):
                    if col.is_0_to_1:
                        col_min, col_max = 100 * col.min_val, 100 * col.max_val
                    else:
                        col_min, col_max = col.min_val, col.max_val

                    fig.update_xaxes(
                        range=[col_min, col_max],
                        row=category_ndx,
                        col=fig_col_ndx,
                        tickvals=[col_min, col_max],
                        showticklabels=show_x_axis,
                    )
                elif isinstance(col, ClassDistributionColumn):
                    fig.update_xaxes(
                        row=category_ndx, col=fig_col_ndx, showticklabels=show_x_axis
                    )
                elif isinstance(col, NumericColumn):
                    fig.update_xaxes(
                        range=[0, 1],
                        row=category_ndx,
                        col=fig_col_ndx,
                        showticklabels=False,
                    )

        fig.update_layout(
            height=height,
            width=self.config["layout_width"],
            barmode="stack",
            plot_bgcolor="rgba(0, 0, 0, 0)",
            paper_bgcolor="rgba(0, 0, 0, 0)",
            font=dict(size=self.config["font_size_data"]),
            yaxis={"autorange": "reversed"},
            margin=go.layout.Margin(
                r=0, b=0, t=20  # right margin  # bottom margin  # top margin
            ),
        )

        # Use low-level plotly interface to update padding / font size
        for a in fig["layout"]["annotations"]:
            # If label for group
            if a["text"] in categories:
                a["x"] = 0.99  # Add padding
                a["font"] = dict(size=self.config["font_size_category"])
            else:
                a["font"] = dict(
                    size=self.config["font_size_heading"]
                )  # Adjust font size for non-category labels

        # Due to a quirk in plotly, need to do some special low-level coding
        # Code from https://community.plotly.com/t/how-to-create-annotated-heatmaps
        # -in-subplots/36686/25
        newfont = [
            go.layout.Annotation(font_size=self.config["font_size_heading"])
        ] * len(fig.layout.annotations)
        fig_annots = [newfont] + [hm.layout.annotations for hm in hms]
        for col_ndx in range(1, len(fig_annots)):
            for k in range(len(fig_annots[col_ndx])):
                coord = coords[col_ndx - 1]
                fig_annots[col_ndx][k]["xref"] = f"x{coord}"
                fig_annots[col_ndx][k]["yref"] = f"y{coord}"
                fig_annots[col_ndx][k]["font_size"] = self.config["font_size_dist"]

        def recursive_extend(mylist, nr):
            # mylist is a list of lists
            result = []
            if nr == 1:
                result.extend(mylist[nr - 1])
            else:
                result.extend(mylist[nr - 1])
                result.extend(recursive_extend(mylist, nr - 1))
            return result

        new_annotations = recursive_extend(fig_annots[::-1], len(fig_annots))
        fig.update_layout(annotations=new_annotations)

        if show_title:
            title = {
                "text": f"{self.dataset_name or ''} {self.model_name or ''} "
                f"Robustness Report",
                "x": 0.5,
                "xanchor": "center",
            }
        else:
            title = None
        fig.update_layout(
            title=title,
            margin=go.layout.Margin(
                r=0, b=0, t=80  # right margin  # bottom margin  # top margin
            ),
        )

        return fig

    def get_color(self, col_ndx):
        return self.config["color_scheme"][col_ndx % len(self.config["color_scheme"])]


def human_format(num):
    num = float("{:.3g}".format(num))
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    return "{}{}".format(
        "{:f}".format(num).rstrip("0").rstrip("."), ["", "K", "M", "B", "T"][magnitude]
    )



"""Taken from https://github.com/tommccoy1/hans/blob/master/templates.py."""

from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.subpopulation import SubpopulationCollection
from robustnessgym.slicebuilders.subpopulations.phrase import HasAnyPhrase


class HansAllPhrases(SubpopulationCollection):
    def __init__(self, *args, **kwargs):
        super(HansAllPhrases, self).__init__(
            subpopulations=HasAnyPhrase.join(
                *[
                    HansSingularNouns(),
                    HansPluralNouns(),
                    HansTransitiveVerbs(),
                    HansPassiveVerbs(),
                    HansIntransitiveVerbs(),
                    HansNPSVerbs(),
                    HansNPZVerbs(),
                    HansPluralNPZVerbs(),
                    HansPrepositions(),
                    HansConjs(),
                    HansPastParticiples(),
                    HansUnderstoodArgumentVerbs(),
                    HansNonEntQuotVerbs(),
                    HansQuestionEmbeddingVerbs(),
                    HansCalledObjects(),
                    HansToldObjects(),
                    HansFoodWords(),
                    HansLocationNounsA(),
                    HansLocationNounsB(),
                    HansWonObjects(),
                    HansReadWroteObjects(),
                    HansAdjectives(),
                    HansAdjectivesCompNonEnt(),
                    HansAdjectivesCompEnt(),
                    HansAdverbs(),
                    HansConstAdv(),
                    HansConstQuotEntailed(),
                    HansRelations(),
                    HansQuestions(),
                    HansNonEntComplementNouns(),
                    HansEntComplementNouns(),
                    HansAdvsNonEntailed(),
                    HansAdvsEntailed(),
                ]
            ),
            *args,
            **kwargs
        )


class HansSingularNouns(HasAnyPhrase):
    def __init__(self):
        super(HansSingularNouns, self).__init__(
            phrase_groups=[
                [
                    "professor",
                    "student",
                    "president",
                    "judge",
                    "senator",
                    "secretary",
                    "doctor",
                    "lawyer",
                    "scientist",
                    "banker",
                    "tourist",
                    "manager",
                    "artist",
                    "author",
                    "actor",
                    "athlete",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansPluralNouns(HasAnyPhrase):
    def __init__(self):
        super(HansPluralNouns, self).__init__(
            phrase_groups=[
                [
                    "professors",
                    "students",
                    "presidents",
                    "judges",
                    "senators",
                    "secretaries",
                    "doctors",
                    "lawyers",
                    "scientists",
                    "bankers",
                    "tourists",
                    "managers",
                    "artists",
                    "authors",
                    "actors",
                    "athletes",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansTransitiveVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansTransitiveVerbs, self).__init__(
            phrase_groups=[
                [
                    "recommended",
                    "called",
                    "helped",
                    "supported",
                    "contacted",
                    "believed",
                    "avoided",
                    "advised",
                    "saw",
                    "stopped",
                    "introduced",
                    "mentioned",
                    "encouraged",
                    "thanked",
                    "recognized",
                    "admired",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansPassiveVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansPassiveVerbs, self).__init__(
            phrase_groups=[
                [
                    "recommended",
                    "helped",
                    "supported",
                    "contacted",
                    "believed",
                    "avoided",
                    "advised",
                    "stopped",
                    "introduced",
                    "mentioned",
                    "encouraged",
                    "thanked",
                    "recognized",
                    "admired",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansIntransitiveVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansIntransitiveVerbs, self).__init__(
            phrase_groups=[
                [
                    "slept",
                    "danced",
                    "ran",
                    "shouted",
                    "resigned",
                    "waited",
                    "arrived",
                    "performed",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansNPSVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansNPSVerbs, self).__init__(
            phrase_groups=[
                [
                    "believed",
                    "knew",
                    "heard",
                    "forgot",
                    "preferred",
                    "claimed",
                    "wanted",
                    "needed",
                    "found",
                    "suggested",
                    "expected",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansNPZVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansNPZVerbs, self).__init__(
            phrase_groups=[["hid", "moved", "presented", "paid", "studied", "stopped"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansPluralNPZVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansPluralNPZVerbs, self).__init__(
            phrase_groups=[
                [
                    "fought",
                    "paid",
                    "changed",
                    "studied",
                    "answered",
                    "stopped",
                    "grew",
                    "moved",
                    "returned",
                    "left",
                    "improved",
                    "lost",
                    "visited",
                    "ate",
                    "played",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansPrepositions(HasAnyPhrase):
    def __init__(self):
        super(HansPrepositions, self).__init__(
            phrase_groups=[["near", "behind", "by", "in front of", "next to"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansConjs(HasAnyPhrase):
    def __init__(self):
        super(HansConjs, self).__init__(
            phrase_groups=[
                ["while", "after", "before", "when", "although", "because", "since"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansPastParticiples(HasAnyPhrase):
    def __init__(self):
        super(HansPastParticiples, self).__init__(
            phrase_groups=[["studied", "paid", "helped", "investigated", "presented"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansUnderstoodArgumentVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansUnderstoodArgumentVerbs, self).__init__(
            phrase_groups=[["paid", "explored", "won", "wrote", "left", "read", "ate"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansNonEntQuotVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansNonEntQuotVerbs, self).__init__(
            phrase_groups=[
                ["hoped", "claimed", "thought", "believed", "said", "assumed"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansQuestionEmbeddingVerbs(HasAnyPhrase):
    def __init__(self):
        super(HansQuestionEmbeddingVerbs, self).__init__(
            phrase_groups=[
                ["wondered", "understood", "knew", "asked", "explained", "realized"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansCalledObjects(HasAnyPhrase):
    def __init__(self):
        super(HansCalledObjects, self).__init__(
            phrase_groups=[["coward", "liar", "hero", "fool"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansToldObjects(HasAnyPhrase):
    def __init__(self):
        super(HansToldObjects, self).__init__(
            phrase_groups=[["story", "lie", "truth", "secret"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansFoodWords(HasAnyPhrase):
    def __init__(self):
        super(HansFoodWords, self).__init__(
            phrase_groups=[
                ["fruit", "salad", "broccoli", "sandwich", "rice", "corn", "ice cream"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansLocationNounsA(HasAnyPhrase):
    def __init__(self):
        super(HansLocationNounsA, self).__init__(
            phrase_groups=[
                [
                    "neighborhood",
                    "region",
                    "country",
                    "town",
                    "valley",
                    "forest",
                    "garden",
                    "museum",
                    "desert",
                    "island",
                    "town",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansLocationNounsB(HasAnyPhrase):
    def __init__(self):
        super(HansLocationNounsB, self).__init__(
            phrase_groups=[["museum", "school", "library", "office", "laboratory"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansWonObjects(HasAnyPhrase):
    def __init__(self):
        super(HansWonObjects, self).__init__(
            phrase_groups=[
                [
                    "race",
                    "contest",
                    "war",
                    "prize",
                    "competition",
                    "election",
                    "battle",
                    "award",
                    "tournament",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansReadWroteObjects(HasAnyPhrase):
    def __init__(self):
        super(HansReadWroteObjects, self).__init__(
            phrase_groups=[
                [
                    "book",
                    "column",
                    "report",
                    "poem",
                    "letter",
                    "novel",
                    "story",
                    "play",
                    "speech",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansAdjectives(HasAnyPhrase):
    def __init__(self):
        super(HansAdjectives, self).__init__(
            phrase_groups=[
                [
                    "important",
                    "popular",
                    "famous",
                    "young",
                    "happy",
                    "helpful",
                    "serious",
                    "angry",
                ]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansAdjectivesCompNonEnt(HasAnyPhrase):
    def __init__(self):
        super(HansAdjectivesCompNonEnt, self).__init__(
            phrase_groups=[["afraid", "sure", "certain"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansAdjectivesCompEnt(HasAnyPhrase):
    def __init__(self):
        super(HansAdjectivesCompEnt, self).__init__(
            phrase_groups=[["sorry", "aware", "glad"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansAdverbs(HasAnyPhrase):
    def __init__(self):
        super(HansAdverbs, self).__init__(
            phrase_groups=[
                ["quickly", "slowly", "happily", "easily", "quietly", "thoughtfully"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansConstAdv(HasAnyPhrase):
    def __init__(self):
        super(HansConstAdv, self).__init__(
            phrase_groups=[
                ["after", "before", "because", "although", "though", "since", "while"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansConstQuotEntailed(HasAnyPhrase):
    def __init__(self):
        super(HansConstQuotEntailed, self).__init__(
            phrase_groups=[["forgot", "learned", "remembered", "knew"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansRelations(HasAnyPhrase):
    def __init__(self):
        super(HansRelations, self).__init__(
            phrase_groups=[["who", "that"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansQuestions(HasAnyPhrase):
    def __init__(self):
        super(HansQuestions, self).__init__(
            phrase_groups=[["why", "how"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansNonEntComplementNouns(HasAnyPhrase):
    def __init__(self):
        super(HansNonEntComplementNouns, self).__init__(
            phrase_groups=[["feeling", "evidence", "idea", "belief"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansEntComplementNouns(HasAnyPhrase):
    def __init__(self):
        super(HansEntComplementNouns, self).__init__(
            phrase_groups=[["fact", "reason", "news", "time"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansAdvsNonEntailed(HasAnyPhrase):
    def __init__(self):
        super(HansAdvsNonEntailed, self).__init__(
            phrase_groups=[["supposedly", "probably", "maybe", "hopefully"]],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )


class HansAdvsEntailed(HasAnyPhrase):
    def __init__(self):
        super(HansAdvsEntailed, self).__init__(
            phrase_groups=[
                ["certainly", "definitely", "clearly", "obviously", "suddenly"]
            ],
            identifiers=[Identifier(_name=self.__class__.__name__)],
        )



"""Transformations using nlpaug."""
from typing import List

try:
    from nlpaug.flow import Pipeline
except ImportError:
    _nlpaug_available = False
    Pipeline = None
else:
    _nlpaug_available = True

from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.transformation import SingleColumnTransformation


class NlpAugTransformation(SingleColumnTransformation):
    """Class for building transformations using nlpaug."""

    def __init__(
        self,
        pipeline: Pipeline,
        num_transformed: int = 1,
        identifiers: List[Identifier] = None,
        *args,
        **kwargs
    ):
        if not _nlpaug_available:
            raise ImportError("Please `pip install nlpaug`.")
        assert isinstance(pipeline, Pipeline), (
            "`pipeline` must be an nlpaug Pipeline object. "
            "Please use \nfrom nlpaug.flow import "
            "Sequential\nrg.NlpAugTransformation(pipeline=Sequential(flow=[...]))."
        )

        # Superclass call
        super(NlpAugTransformation, self).__init__(
            num_transformed=num_transformed,
            identifiers=Identifier.range(
                n=num_transformed,
                _name=self.__class__.__name__,
                pipeline=[
                    Identifier(
                        _name=augmenter.name,
                        src=augmenter.aug_src
                        if hasattr(augmenter, "aug_src")
                        else None,
                        action=augmenter.action,
                        method=augmenter.method,
                    )
                    for augmenter in pipeline
                ],
            )
            if not identifiers
            else identifiers,
            *args,
            **kwargs
        )

        # Set the pipeline
        self._pipeline = pipeline

    @property
    def pipeline(self):
        return self._pipeline

    def single_column_apply(self, column_batch: List[str]) -> List[List[str]]:
        # Apply the nlpaug pipeline
        augmented_texts = self.pipeline.augment(
            data=column_batch,
            n=self.num_transformed,
        )

        if self.num_transformed == 1:
            augmented_texts = [augmented_texts]

        # Transpose the list of lists from [4 x 32] to [32 x 4] and return
        return list(map(list, zip(*augmented_texts)))



"""Cachedop with Spacy."""
import json
from typing import List

import cytoolz as tz
import torch

try:
    import spacy
    from spacy.tokens import Doc
except ImportError:
    _spacy_available = False
    Doc = None
else:
    _spacy_available = True

from robustnessgym.core.cachedops import SingleColumnCachedOperation
from robustnessgym.core.dataset import BatchOrDataset


class Spacy(SingleColumnCachedOperation):
    """Class for running the Spacy pipeline using a CachedOperation."""

    def __init__(
        self,
        lang: str = "en_core_web_sm",
        nlp: "spacy.language.Language" = None,
        neuralcoref: bool = False,
        device: str = None,
        *args,
        **kwargs
    ):
        if not _spacy_available:
            raise ImportError("Please `pip install spacy`.")

        # Set all the parameters
        self.lang = lang
        self.neuralcoref = neuralcoref
        self._prebuilt = True

        # Set the device
        self._on_gpu = False
        if device and (device == "gpu" or device.startswith("cuda")):
            spacy.prefer_gpu(
                gpu_id=0 if ":" not in device else int(device.split(":")[1])
            )
            # Spacy sets the default torch float Tensor to torch.cuda.FloatTensor,
            # which causes other GPU cachedops to crash.
            torch.set_default_tensor_type("torch.FloatTensor")
            self._on_gpu = True

        # Load up the Spacy module
        self._nlp = nlp
        if not nlp:
            self._nlp = self._load_spacy(lang=lang)
            self._prebuilt = False

        # Add neuralcoref
        self._add_neuralcoref()

        if not nlp:
            super(Spacy, self).__init__(
                lang=lang,
                neuralcoref=neuralcoref,
                *args,
                **kwargs,
            )
        else:
            super(Spacy, self).__init__(
                lang=nlp.lang,
                # No need to pass in neuralcoref separately, it's already in the
                # pipeline if neuralcoref=True
                pipeline=nlp.pipe_names,
                *args,
                **kwargs,
            )
            print(
                "Warning: Spacy.encode does not support arbitrary nlp pipelines so "
                "information stored in the Doc object may be lost in encoding."
            )

    @staticmethod
    def _load_spacy(lang: str = "en_core_web_sm"):
        """Load the Spacy nlp pipeline."""
        return spacy.load(lang)

    def _add_neuralcoref(self):
        """Add the neuralcoref pipeline to Spacy."""
        if self.neuralcoref:
            try:
                import neuralcoref as nc

                nc.add_to_pipe(self.nlp)
            except ImportError:
                print(
                    "Can't import neuralcoref. Please install neuralcoref using:\n"
                    "git clone https://github.com/huggingface/neuralcoref.git\n"
                    "cd neuralcoref\n"
                    "pip install -r requirements.txt\n"
                    "pip install -e ."
                )

    def __call__(
        self,
        batch_or_dataset: BatchOrDataset,
        columns: List[str],
        batch_size: int = 8192,
    ) -> BatchOrDataset:
        return super().__call__(batch_or_dataset, columns, batch_size)

    @property
    def nlp(self):
        """Return the nlp pipeline."""
        return self._nlp

    @classmethod
    def encode(cls, obj: Doc) -> str:
        """Encode the Doc object.

        Args:
            obj:

        Returns:
        """
        # JSON dump the Doc
        doc_json = obj.to_json()

        if obj._.has("huggingface_neuralcoref"):
            # Create a helper function that turns a Span into a dictionary
            span_to_dict = lambda span: {
                "start": span.start,
                "end": span.end,
                "text": span.text,
            }

            # Create a helper function that converts a Cluster (output of
            # neuralcoref) into a dictionary
            cluster_to_dict = lambda cluster: {
                "i": cluster.i,
                "main": span_to_dict(cluster.main),
                "mentions": [span_to_dict(span) for span in cluster.mentions],
            }

            # Apply the helper functions to construct a dictionary for the
            # neuralcoref information
            neuralcoref_dict = {
                "neuralcoref": [
                    cluster_to_dict(cluster) for cluster in obj._.coref_clusters
                ]
            }

            # Combine the neuralcoref dictionary with the doc_json
            doc_json = tz.merge(doc_json, neuralcoref_dict)

        # Convert the Spacy Doc to json before caching
        return json.dumps(doc_json)

    def single_column_apply(self, column_batch: List, *args, **kwargs) -> List:
        """Apply to a single column.

        Args:
            column_batch:
            *args:
            **kwargs:

        Returns:
        """
        if self._on_gpu:
            # Adjust the default Tensor type: this is instantaneous
            torch.set_default_tensor_type("torch.cuda.FloatTensor")

        # Apply Spacy's pipe method to process the examples
        docs = list(self.nlp.pipe(column_batch))

        if self._on_gpu:
            # Reset the default Tensor type: this is instantaneous
            torch.set_default_tensor_type("torch.FloatTensor")

        return docs

    @classmethod
    def tokens(cls, decoded_batch: List) -> List[List[str]]:
        """For each example, returns the list of tokens extracted by Spacy for
        each key.

        Spacy stores the span of each token under the "tokens" key. This
        function extracts the tokens from the text using the span of
        each token.
        """

        token_batch = []
        # Iterate over each decoded Doc dictionary
        for doc_dict in decoded_batch:
            tokens = []
            for token_dict in doc_dict["tokens"]:
                tokens.append(doc_dict["text"][token_dict["start"] : token_dict["end"]])

            token_batch.append(tokens)

        return token_batch

    @classmethod
    def entities(cls, decoded_batch: List) -> List[List[dict]]:
        """For each example, returns the list of entity extracted by Spacy for
        each column."""
        return [doc_dict["ents"] for doc_dict in decoded_batch]

    @classmethod
    def sentences(cls, decoded_batch: List) -> List[List[str]]:
        """For each example, returns the list of sentences extracted by Spacy
        for each column."""
        return [
            [
                doc_dict["text"][sent["start"] : sent["end"]]
                for sent in doc_dict["sents"]
            ]
            for doc_dict in decoded_batch
        ]

    @classmethod
    def num_tokens(cls, decoded_batch: List) -> List[int]:
        """For each example, returns the length or the number of tokens
        extracted by Spacy for each column."""
        return [len(doc_dict["tokens"]) for doc_dict in decoded_batch]



import re
import statistics
from typing import Callable, Sequence, Union

try:
    import nltk
except ImportError:
    _nltk_available = False
else:
    _nltk_available = True
import numpy as np
import torch

try:
    from rouge_score import rouge_scorer
except ImportError:
    _rouge_score_available = False
    rouge_scorer = None
else:
    _rouge_score_available = True
from sklearn.metrics import accuracy_score, f1_score


def get_metric(name: str) -> Callable:
    """Get metrics from string names."""
    if name == "accuracy":
        return accuracy
    elif name == "f1":
        return f1
    elif name == "f1_micro":
        return f1_micro
    elif name == "f1_macro":
        return f1_macro
    else:
        raise NotImplementedError(f"Metric name {name} not recognized.")


def accuracy(
    predictions: Union[list, np.array, torch.Tensor],
    labels: Union[list, np.array, torch.Tensor],
):
    """Calculate accuracy."""
    return accuracy_score(y_true=labels, y_pred=predictions)


def f1(
    predictions: Union[list, np.array, torch.Tensor],
    labels: Union[list, np.array, torch.Tensor],
):
    """Calculate F1 score for binary classification."""
    return f1_score(y_true=labels, y_pred=predictions)


def f1_micro(
    predictions: Union[list, np.array, torch.Tensor],
    labels: Union[list, np.array, torch.Tensor],
):
    """Calculate micro F1 score for multi-class classification."""
    return f1_score(y_true=labels, y_pred=predictions, average="micro")


def f1_macro(
    predictions: Union[list, np.array, torch.Tensor],
    labels: Union[list, np.array, torch.Tensor],
):
    """Calculate macro F1 score for multi-class classification."""
    return f1_score(y_true=labels, y_pred=predictions, average="macro")


def class_distribution(
    labels: Union[list, np.array, torch.Tensor],
    num_classes: int = None,
    min_label: int = 0,
):
    """Calculate the aggregated class distribution."""
    if isinstance(labels, list):
        labels = np.array(labels)

    if len(labels.shape) == 1:
        # Find the unique labels and their counts
        unique_labels, counts = np.unique(labels, return_counts=True)

        # Calculate the number of classes
        if num_classes is None:
            max_label = np.max(unique_labels)
            num_classes = max_label - min_label + 1

        # Fill out the distribution
        dist = np.zeros(num_classes)
        dist[(unique_labels - min_label).astype(int)] = counts / labels.shape[0]
        return dist
    elif len(labels.shape) == 2:
        return np.mean(labels, axis=0)
    else:
        raise ValueError("`labels` must be 1 or 2-dimensional.")


# TODO Refactor into separate class for each metric
# TODO change signature of compute_metric
def compute_metric(
    metric: str,
    predictions: Union[Sequence, torch.Tensor],
    labels: Union[Sequence, torch.Tensor],
    num_classes: int,
):
    """Compute metric given predictions and target labels
    Args:
        metric: name of metric
        predictions: A sequence of predictions (rouge metrics) or a torch Tensor
        (other metrics) containing predictions
        labels: A sequence of labels (rouge metrics) or a torch Tensor (other metrics)
        containing target labels
        num_classes: number of classes
    """
    if isinstance(predictions, torch.Tensor):
        predictions = predictions.cpu()
    if isinstance(labels, torch.Tensor):
        labels = labels.cpu()

    # Classification metrics
    if metric == "accuracy":
        return accuracy(predictions=predictions, labels=labels)
    elif metric == "f1":
        return f1(predictions=predictions, labels=labels)
    elif metric == "f1_micro":
        return f1_micro(predictions=predictions, labels=labels)
    elif metric == "f1_macro":
        return f1_macro(predictions=predictions, labels=labels)

    elif metric in ("Rouge-1", "Rouge-2", "Rouge-L"):
        # Calculate rouge scores
        if metric == "Rouge-1":
            metric_id = "rouge1"
        elif metric == "Rouge-2":
            metric_id = "rouge2"
        else:
            metric_id = "rougeLsum"
        scorer = rouge_scorer.RougeScorer([metric_id], use_stemmer=True)
        # TODO Remove summarizaton-specific 'format_summary' call
        # TODO Don't call scorer.score separately for each metric
        score = statistics.mean(
            scorer.score(format_summary(reference), format_summary(pred))[
                metric
            ].fmeasure
            for reference, pred in zip(labels, predictions)
        )

    elif metric == "class_dist":
        # Calculate class distribution
        score = class_distribution(labels=labels, num_classes=num_classes)
    elif metric == "pred_dist":
        # Calculate predicted class distribution
        score = class_distribution(labels=predictions, num_classes=num_classes)
    else:
        raise NotImplementedError

    return score


def format_summary(x: str) -> str:
    """Format summary text for computing rouge."""
    re.sub("<n>", "", x)  # remove pegasus newline char
    return "\n".join(nltk.sent_tokenize(x))



from copy import deepcopy
from typing import Callable, Dict, List


def singlecolumn(func: Callable):
    """Assert that func is called with a single column.

    Mainly used with .apply(..) methods for CachedOperation and SliceBuilder.

    Args:
        func: function to wrap

    Returns: decorated function
    """

    def _singlecolumn(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ):
        assert len(columns) == 1, "Must pass in a single column."
        return func(self, batch, columns, *args, **kwargs)

    return _singlecolumn


def prerequisites(*args):
    """
    Decorator to add a prerequisites attribute to any class.
    Args:
        *args: list of prerequisites

    Returns: a decorator
    """

    def _decorator(cls):
        _old_init = deepcopy(cls.__init__)

        def _new_init(self, *_args, **kwargs):
            _old_init(self, *_args, **kwargs)
            self.prerequisites = set(args)

        cls.__init__ = _new_init

        return cls

    return _decorator


def function_register():
    registry = {}

    def registrar(func):
        # Register the function
        registry[func.__name__] = func

        # Mark the fact that the function is decorated
        func.decorator = registrar

        return func

    registrar.all = registry
    return registrar


# Create processors that keep track of batch and dataset operations
batch_processing = function_register()
dataset_processing = function_register()


def methods_with_decorator(cls, decorator):
    """Returns all methods in cls with decorator as the outermost decorator.

    Credit: https://stackoverflow.com/questions/5910703/how-to-get-all-methods-of-a
    -python-class-with-given-decorator
    """
    for maybe_decorated in cls.__dict__.values():
        if hasattr(maybe_decorated, "decorator"):
            if maybe_decorated.decorator == decorator:
                yield maybe_decorated



from functools import reduce
from typing import Dict, List

import torch
import torch.nn as nn

from robustnessgym import Dataset
from robustnessgym.core.cachedops import SingleColumnCachedOperation
from robustnessgym.core.decorators import singlecolumn


class ActivationExtractor:
    """Class for extracting activations of a targetted intermediate layer."""

    def __init__(self):
        self.activation = None

    def add_hook(self, module, input, output):
        self.activation = output


class ActivationCachedOp(SingleColumnCachedOperation):
    def __init__(self, model: nn.Module, target_module: str, device: int = None):
        """ A cached operation that runs a forward pass over 
        each example in the dataset and stores model activations in a new column. 
        TODO: test on an NLP model 

        Args:
            model (nn.Module): the torch model from which activations are extracted
            target_module (str): the name of the submodule of `model` (i.e. an
                intermediate layer) that outputs the activations we'd like to extract.
                For nested submodules, specify a path separated by "." (e.g.
                `ActivationCachedOp(model, "block4.conv")`).
            device (int, optional): the device for the forward pass. Defaults to None,
                in which case the CPU is used.
        """
        self.model = model
        self.device = device

        try:
            target_module = _nested_getattr(model, target_module)
        except nn.modules.module.ModuleAttributeError:
            raise ValueError(f"model does not have a submodule {target_module}")

        self.extractor = ActivationExtractor()
        target_module.register_forward_hook(self.extractor.add_hook)

        super(ActivationCachedOp, self).__init__()

    # TODO(sabri): make a proper encoder/decoder pair using torch.save
    @classmethod
    def encode(cls, obj: torch.Tensor) -> torch.Tensor:
        return obj

    @classmethod
    def decode(cls, obj: torch.Tensor) -> torch.Tensor:
        return obj

    def prepare_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs,
    ) -> None:

        # First reset the scores
        if self.device is not None:
            self.model.to(self.device)

        # Prepare the dataset
        super(ActivationCachedOp, self).prepare_dataset(
            dataset=dataset,
            columns=columns,
            batch_size=batch_size,
            *args,
            **kwargs,
        )

    @singlecolumn
    def apply(self, batch: Dict[str, List], columns: List[str], **kwargs) -> List:
        inputs = batch[columns[0]]

        if self.device is not None:
            inputs = inputs.to(self.device)

        self.model(inputs)
        return self.extractor.activation.cpu().detach()


def _nested_getattr(obj, attr, *args):
    """Get a nested property from an object.

    Example:
    ```
        model = ...
        weights = _nested_getattr(model, "layer4.weights")
    ```
    """
    return reduce(lambda o, a: getattr(o, a, *args), [obj] + attr.split("."))



from typing import Callable, Dict, List, Tuple, Union

import numpy as np

from robustnessgym.core.cachedops import ScoreOperation
from robustnessgym.core.dataset import Batch, Dataset
from robustnessgym.core.identifier import Identifier
from robustnessgym.slicebuilders.subpopulation import Subpopulation


class BinningMixin:
    def __init__(
        self,
        intervals: List[Tuple[Union[int, float, str], Union[int, float, str]]],
        bin_creation_fn: Callable = None,
        bin_fn: Callable = None,
        *args,
        **kwargs
    ):
        super(BinningMixin, self).__init__(*args, **kwargs)

        # Set the intervals
        self.intervals = intervals
        self.left_limits = None
        self.right_limits = None

        # Keep track of scores
        self.scores = []

        # Assign the bin fns
        if bin_creation_fn and bin_fn:
            self.create_bins = bin_creation_fn
            self.bin = bin_fn

    def _reset_scores(self):
        self.scores = []

    def replace_percentile(self, limit):
        if isinstance(limit, str) and limit.endswith("%"):
            return np.percentile(self.scores, float(limit.replace("%", "")))
        elif isinstance(limit, float) or isinstance(limit, int):
            return limit
        else:
            raise NotImplementedError

    def create_bins(self):
        for i in range(len(self.intervals)):
            (left_limit, right_limit) = self.intervals[i]
            self.intervals[i] = (
                self.replace_percentile(left_limit),
                self.replace_percentile(right_limit),
            )

        self.left_limits = np.array([interval[0] for interval in self.intervals])
        self.right_limits = np.array([interval[1] for interval in self.intervals])

    def bin(self, scores: List[Union[int, float]]) -> np.ndarray:
        # Convert to np.ndarry
        scores = np.array(scores)

        # Bin the scores
        return (
            (self.left_limits <= scores[:, np.newaxis])
            & (scores[:, np.newaxis] <= self.right_limits)
        ).astype(np.int32)


class ScoreSubpopulation(Subpopulation, BinningMixin):
    def __init__(
        self,
        intervals: List[Tuple[Union[int, float, str], Union[int, float, str]]],
        identifiers: List[Identifier] = None,
        score_fn: Callable = None,
        bin_creation_fn: Callable = None,
        bin_fn: Callable = None,
        *args,
        **kwargs
    ):

        if not identifiers:
            if score_fn:
                identifiers = [
                    Identifier(
                        _name=self.__class__.__name__,
                        gte=interval[0],
                        lte=interval[1],
                        score_fn=score_fn,
                    )
                    for interval in intervals
                ]
            else:
                identifiers = [
                    Identifier(
                        _name=self.__class__.__name__, gte=interval[0], lte=interval[1]
                    )
                    for interval in intervals
                ]

        super(ScoreSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=identifiers,
            bin_creation_fn=bin_creation_fn,
            bin_fn=bin_fn,
            *args,
            **kwargs,
        )

        # Assign the score fn
        if score_fn:
            self.score = score_fn

    def prepare_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs
    ) -> None:

        # First reset the scores
        self._reset_scores()

        # Prepare the dataset
        super(ScoreSubpopulation, self).prepare_dataset(
            dataset=dataset,
            columns=columns,
            batch_size=batch_size,
            *args,
            **kwargs,
        )

        # Create the bins
        self.create_bins()

    def prepare_batch(self, batch: Batch, columns: List[str], *args, **kwargs) -> None:

        # Compute the scores
        if isinstance(self.score, ScoreOperation):
            self.scores.extend(self.score.retrieve(batch=batch, columns=columns))
        elif isinstance(self.score, Callable):
            self.scores.extend(self.score(batch=batch, columns=columns))
        else:
            raise RuntimeError("score function invalid.")

    def score(
        self, batch: Dict[str, List], columns: List[str], *args, **kwargs
    ) -> np.ndarray:
        raise NotImplementedError("Return a vector of float scores for each example.")

    def apply(
        self,
        slice_membership: np.ndarray,
        batch: Dict[str, List],
        columns: List[str],
        *args,
        **kwargs
    ) -> np.ndarray:
        # Keep track of the score of each example
        if isinstance(self.score, ScoreOperation):
            scores = self.score.retrieve(batch=batch, columns=columns)
        elif isinstance(self.score, Callable):
            scores = self.score(batch=batch, columns=columns)
        else:
            raise RuntimeError("score function invalid.")

        assert (
            len(scores) == slice_membership.shape[0]
        ), "Must have exactly one score per example."

        return self.bin(scores=scores)


class MultiScoreSubpopulation(Subpopulation, BinningMixin):
    def __init__(
        self,
        intervals: List[Tuple[Union[int, float, str], Union[int, float, str]]],
        identifiers: List[Identifier] = None,
        score_fn: Callable = None,
        bin_creation_fn: Callable = None,
        bin_fn: Callable = None,
        *args,
        **kwargs
    ):

        if not identifiers:
            if score_fn:
                identifiers = [
                    Identifier(
                        _name=self.__class__.__name__,
                        gte=interval[0],
                        lte=interval[1],
                        score_fn=score_fn,
                    )
                    for interval in intervals
                ]
            else:
                identifiers = [
                    Identifier(
                        _name=self.__class__.__name__, gte=interval[0], lte=interval[1]
                    )
                    for interval in intervals
                ]

        super(MultiScoreSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=identifiers,
            bin_creation_fn=bin_creation_fn,
            bin_fn=bin_fn,
            *args,
            **kwargs,
        )

        # Assign the score fn
        if score_fn:
            self.score = score_fn

    def prepare_dataset(
        self,
        dataset: Dataset,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs
    ) -> None:

        # First reset the scores
        self._reset_scores()

        # Prepare the dataset
        super(MultiScoreSubpopulation, self).prepare_dataset(
            dataset=dataset,
            columns=columns,
            batch_size=batch_size,
        )

        # Create the bins
        self.create_bins()

    def prepare_batch(self, batch: Batch, columns: List[str], *args, **kwargs) -> None:

        # Compute the scores
        if isinstance(self.score, ScoreOperation):
            self.scores.extend(self.score.retrieve(batch=batch, columns=columns))
        elif isinstance(self.score, Callable):
            self.scores.extend(self.score(batch=batch, columns=columns))
        else:
            raise RuntimeError("score function invalid.")



"""Identifiers for objects in Robustness Gym."""
from __future__ import annotations

import ast
import json
from typing import Any, Callable, List, Union

# from robustnessgym.core.tools import persistent_hash


class Identifier:
    """Class for creating identifiers for objects in Robustness Gym."""

    def __init__(
        self,
        _name: str,
        _index: Union[str, int] = None,
        **kwargs,
    ):

        self._name = _name
        self._index = str(_index) if _index is not None else None
        self._parameters = kwargs

        # Add the parameter
        for param, value in self.parameters.items():
            self.add_parameter(param, value)

    @property
    def name(self):
        """Base name."""
        return self._name

    @property
    def index(self):
        """Index associated with the identifier."""
        return self._index

    @property
    def parameters(self):
        """Additional parameters contained in the identifier."""
        return self._parameters

    @classmethod
    def range(cls, n: int, _name: str, **kwargs) -> List[Identifier]:
        """Create a list of identifiers, with index varying from 1 to `n`."""

        if n > 1:
            return [cls(_name=_name, _index=i, **kwargs) for i in range(1, n + 1)]
        return [cls(_name=_name, **kwargs)]

    def __call__(self, **kwargs):
        """Call the identifier with additional parameters to return a new
        identifier."""
        ident = Identifier.loads(self.dumps())
        for parameter, value in kwargs.items():
            ident.add_parameter(parameter, value)
        return ident

    def __repr__(self):
        params = ", ".join([f"{k}={v}" for k, v in self.parameters.items()])
        if self.index is not None:
            return (
                f"{self.name}-{self.index}({params})"
                if len(params) > 0
                else f"{self.name}-{self.index}"
            )
        return f"{self.name}({params})" if len(params) > 0 else f"{self.name}"

    def __hash__(self):
        # return persistent_hash(str(self))
        return hash(str(self))

    def __eq__(self, other: Union[Identifier, str]):
        return str(self) == str(other)

    def dumps(self):
        """Dump the identifier to JSON."""
        return json.dumps(self.__dict__)

    @staticmethod
    def _parse_args(s: str):
        """https://stackoverflow.com/questions/49723047/parsing-a-string-as-a-
        python-argument-list."""
        args = "f({})".format(s)
        tree = ast.parse(args)
        funccall = tree.body[0].value
        # return {arg.arg: ast.literal_eval(arg.value) for arg in funccall.keywords}
        params = {}
        for arg in funccall.keywords:
            try:
                params[arg.arg] = ast.literal_eval(arg.value)
            except ValueError:
                params[arg.arg] = arg.value.id
        return params

    @classmethod
    def parse(cls, s: str) -> Identifier:
        """Parse in an identifier from string."""
        # Parse out the various components
        if "(" in s:
            name_index, params = s.split("(")
            params = params.split(")")[0]
        else:
            name_index = s
            params = None

        # Create the name and index
        if "-" in name_index:
            name, index = name_index.split("-")[:-1], name_index.split("-")[-1]
            name = "-".join(name)
            if index.isnumeric():
                index = int(index)
            else:
                name = "-".join([name, index])
                index = None
        else:
            name = name_index
            index = None

        # Parse out the params
        if params is not None:
            params = cls._parse_args(params)
        else:
            params = {}

        return cls(_name=name, _index=index, **params)

    def without(self, *params) -> Identifier:
        """Returns an identifier without `params`."""
        return Identifier(
            self.name,
            self.index,
            **{k: v for k, v in self.parameters.items() if k not in set(params)},
        )

    @classmethod
    def loads(cls, s: str):
        """Load the identifier from JSON."""
        identifier = Identifier(_name="")
        identifier.__dict__ = json.loads(s)
        return identifier

    def add_parameter(self, parameter: str, value: Any) -> None:
        """Add a parameter to the identifier."""
        if isinstance(value, Callable):
            self.parameters[parameter] = ".".join(
                [str(value.__module__), str(value.__name__)]
            )
        else:
            self.parameters[parameter] = value


# Assign Id as an alias for the Identifier class
Id = Identifier



import logging
import os
import tempfile
import uuid
from datetime import datetime
from typing import Union

logger = logging.getLogger(__name__)


def initialize_logging(
    log_dir: str = tempfile.gettempdir(),
    log_name: str = "robustnessgym.log",
    format: str = "[%(asctime)s][%(levelname)s][%(name)s:%(lineno)s] :: %(message)s",
    level: int = logging.WARNING,
) -> None:
    """Initialize logging for Robustness Gym."""

    # Generate a new directory using the log_dir, if it doesn't exist
    date = datetime.now().strftime("%Y_%m_%d")
    time = datetime.now().strftime("%H_%M_%S")
    uid = str(uuid.uuid4())[:8]
    log_path = os.path.join(log_dir, date, time, uid)

    # Make the logdir
    os.makedirs(log_path, exist_ok=True)

    # Initialize logging
    logging.basicConfig(
        format=format,
        level=level,
        handlers=[
            logging.FileHandler(os.path.join(log_path, log_name)),
            logging.StreamHandler(),
        ],
    )

    # Set logging levels for dependencies
    set_logging_level_for_imports()

    logger.info("Logging initialized.")


def set_logging_level_for_imports(level: int = logging.WARNING) -> None:
    """Set logging levels for dependencies."""
    # Set levels for imports
    logging.getLogger("tensorflow").setLevel(level)
    logging.getLogger("matplotlib").setLevel(level)
    logging.getLogger("textattack").setLevel(level)
    logging.getLogger("filelock").setLevel(level)


def set_logging_level(level: Union[int, str] = logging.INFO):
    """Set logging level for Robustness Gym."""
    # Set the top-level logger
    if isinstance(level, int):
        logging.getLogger("robustnessgym").setLevel(level)
    elif isinstance(level, str):
        logging.getLogger("robustnessgym").setLevel(
            {
                "debug": logging.DEBUG,
                "info": logging.INFO,
                "warning": logging.WARNING,
                "critical": logging.CRITICAL,
                "fatal": logging.FATAL,
            }[level]
        )
    else:
        raise NotImplementedError(f"Level `{level}` not recognized.")
